! am 3/20/06 copied from /net/suncat/e/maute/plasmasphere/ctip_apex/src_coup
!   dynamo_v1.8.F and renamed file since this is the dynamo if 
!   field-line integral from GIP are used
!
!
      module dynamo_gip_module
!----------------------------------------------------------------------- 
! BOP
! !MODULE: dynamo_gip_module
! !DESCRIPTION: 
!
! Module for electrodynamic dynamo (the "e" in tiegcm). 
!
! 5/02 btf: This is the "new" dynamo code, adapted by Astrid Maute
! and Art Richmond from the original ("old") dynamo in tgcm15,
! written by E. Cicely Ridley completed in about 1992. 
!
! Sub prep_dynamo is called from advance. Prep_dynamo prepares dynamo 
! input fields and gathers them to the root task (mp_dynamo_gather).
! Next, advance calls sub dynamo (the dynamo driver), which controls
! the dynamo subroutine calls. Prep_dynamo and the dynamo driver are
! called once per timestep, after sub dynamics.
!
! Since the dynamo is serial (as of 4/02), the master task must receive
! subdomain data from other tasks for input fields to the dynamo 
! (on the geographic grid). These fields are exchanged via sub 
! mp_dynamo_gather which is called from prep_dynamo below.
!
! !USES:
!
      use params_module,only: 
     |  nlon,   ! number of geographic longitudes (at 5 deg, nlon==72)
     |  nlonp1, ! nlon+1
     |  nlonp4, ! nlon+4
     |  nlat,   ! number of geographic latitudes (at 5 deg, nlat==36)
     |  nlatp1, ! nlat+1
     |  nlev,   ! number of pressure levels (e.g., 28)
     |  nlevp1,  ! nlev+1; nlev number of pressure levels (e.g., 28)
     |  nmlon,  ! number of geomagnetic grid longitudes
     |  nmlonp1,! nmlon+1
     |  nmlat,  ! number of geomagnetic grid latitudes
     |  nmlatp1,! nmlat+1
     |  nmlath, ! (nmlat+1)/2 (index to magnetic equator)
     |  nmlev   ! number of geomagnetic pressure levels (nmlev==nlevp1+3)
      use dynamo_module,only: pfrac,phihm
      use addfld_module,only: addfld
      
      implicit none
!
! Dimensions of the 5 grid resolutions for the multi-grid PDE:
      integer,parameter ::
     |  nmlon0=nmlon+1,
     |  nmlat0=(nmlat+1)/2,
     |  nmlon1=(nmlon0+1)/2,
     |  nmlat1=(nmlat0+1)/2,
     |  nmlon2=(nmlon1+1)/2,
     |  nmlat2=(nmlat1+1)/2,
     |  nmlon3=(nmlon2+1)/2,
     |  nmlat3=(nmlat2+1)/2,
     |  nmlon4=(nmlon3+1)/2,
     |  nmlat4=(nmlat3+1)/2
!
! Space needed for descretized coefficients of dynamo pde at all 
!   5 levels of resolution:
!
      integer,parameter ::
     |  ncee=10*nmlon0*nmlat0+9*(nmlon1*nmlat1+nmlon2*nmlat2+nmlon3*
     |    nmlat3+nmlon4*nmlat4)
!
! Coefficients are stored in 1-d array cee(ncee)
! cee transmits descretized dynamo PDE coefficients to the multi-grid 
!   mudpack solver. (cee was formerly in ceee.h)
! The common block /cee_com/ is retained from earlier versions because
!   of the equivalencing below of coefficient arrays c0, c1, etc.
!
      real :: cee_gip(ncee) 
      common/cee_com_gip/ cee_gip
!
! The following parameters nc0,nc1,... are pointers to the beginning of 
!   the coefficients for each level of resolution.
!
      integer,parameter ::
     | nc0=1,
     | nc1=nc0+10*nmlon0*nmlat0,
     | nc2=nc1+9 *nmlon1*nmlat1,
     | nc3=nc2+9 *nmlon2*nmlat2,
     | nc4=nc3+9 *nmlon3*nmlat3
!
! nc(1:6) are pointers to beginning of coefficient blocks at each of 
!   5 levels of resolution: 
! nc(1) = nc0, pointer to coefficients for highest resolution.
! nc(2) = nc1, pointer to coefficients at half the resolution of nc0, 
!   and so on for nc(3), nc(4), nc(5), etc.
! nc(6) = ncee, the dimension of the entire cee array, containing
!   coefficients for all 5 levels of resolution.
! 
      integer :: nc_gip(6)
!
      real :: 
     |  c0(nmlon0,nmlat0,10),
     |  c1(nmlon1,nmlat1,9),
     |  c2(nmlon2,nmlat2,9),
     |  c3(nmlon3,nmlat3,9),
     |  c4(nmlon4,nmlat4,9)
      equivalence 
     |  (cee_gip,c0), 
     |  (cee_gip(nc1),c1),
     |  (cee_gip(nc2),c2),
     |  (cee_gip(nc3),c3),
     |  (cee_gip(nc4),c4)
!
! Unmodified coefficients for using modified mudpack:
      real,dimension(nmlon0,nmlat0,9) :: cofum_gip
!
! phim: Single level dynamo potential in geomagnetic coordinates, as
!   output by PDE solver mud (formerly in dynphi.h):
! The dynamo is symmetric about the magnetic equator, but the high latitude
!  is anti-symmetric in both hemispheres.  However, since Mudpack uses the
!  NH potential pattern, then the SH potential pattern must be added
!  back into the 2-D phim before the call threed, and before it is
!  transformed to geographic coordinates.
! jn index for NH part of potential (nmlat down to ~nmlat0)
! jp index for NH pfrac (nmlat0 down to 1), the fraction of the dynamo
!
      real,dimension(nmlonp1,nmlat) :: phim
      integer :: jn,jp
      
! Inputs to the dynamo on the geographic grid. (formerly in dynamo.h):
      real,dimension(nlonp4,nlat,nlevp1) ::
     |  zpoten           ! geopotential height
     
! Fields on geomagnetic grid (at a single latitude):
! (suffix "m" is for "mag")
      real,dimension(nmlonp1,-2:nlevp1) :: 
     |  zpotenm           ! geopotential height
!      
! electric field
      real,dimension(nmlonp1,nmlat) :: ed1,ed2
      real,dimension(nmlonp1,nmlat,-2:nlevp1) :: 
     |  ed13d, ed23d
!
! Coefficients and RHS terms for PDE on geomagnetic grid:
! (formerly in coefm.h)
!
      real,dimension(nmlonp1,nmlat) ::
     |  zigm11,  ! sigma11*cos(theta0)
     |  zigmc,   ! sigmac
     |  zigm2,   ! sigma2
     |  zigm22   ! sigma22/cos(theta0)
!
! rim(1)=id(1), rim(2)=id(2)/cos(theta0)
      real,dimension(nmlonp1,nmlat,2) :: rim 
      real,dimension(nmlonp1,nmlath)  :: rhs ! right-hand side 
!
! 3-d geopotential heights in geomagnetic space (was zzm in dynphi.h)
      real,dimension(nmlonp1,nmlat,-2:nlevp1) :: zpotenm3d
!
! isolve = 0 -> original mud version 5.
! isolve = 1 -> muh hybrid solver (only as direct solver -- slow)
! isolve = 2 -> modified mudpack solver (mudmod)
! Default is isolve=2 for DYNAMO=2 (new dynamo), however under SGI/IRIX,
!   the hybrid solver is used because SGI crashes in mudmod if isolve==2. 
!
#ifdef IRIX
      integer,parameter :: isolve = 1
#else
      integer,parameter :: isolve = 2 ! default is 2 for new dynamo
!      integer,parameter :: isolve = 1 
!      integer,parameter :: isolve = 0
#endif
! 
! coupled dynamo with GIP
      character(len=3), parameter :: which_dynamo = 'GIP'
!
! am 10/04 define potential model : NONE or HEELIS - first test NONE
!      character(len=*),parameter :: potential_model='NONE'
!      character(len=*),parameter :: potential_model='HEELIS'
!
! Electric potential from heelis or weimer:
!      real,dimension(nmlonp1,nmlat0) :: pfrac  ! NH fraction of potential
!      real,dimension(nmlonp1,nmlat)  :: phihm  ! potential in magnetic
!      
! Work array (private to this module)
      real :: wkarray(-15:nmlon0+16,nmlat0)
      
! For dot products:
      real,parameter :: unitvm(nmlon)=1., unitv(nlon)=1.
!
! !REVISION HISTORY:
! 24.06.05  Astrid Maute: merge TIEGCM version1.8 into coupling with CTIP
!           for George Millward
! 05.02.04  Astrid Maute: include headers
! 
! comments from the original coupled code with CTIP
! am 10/04: adapte the electrodynamo from tiegcm1.6 for the
! coupling with a plasmasphere model
! -electro-dynamo is serial
! -high latitude potential is still in the code but shouldn't be 
! specified (set crit(1) and crit(2) to zero or small value
! crit(1), crit(2) colatitude for boundary of specified high lat. 
! potential (lat_m < crit(1) ) and transitions zone (crit(1)< lat_m < crit(2)
! - output eletric potential in magnetic coordinates
!          electric field in magnetic coordinates
!
! Module for electrodynamic dynamo (the "e" in tiegcm). 
!
! 5/02 btf: This is the "new" dynamo code, adapted by Astrid Maute
! and Art Richmond from the original ("old") dynamo in tgcm15,
! written by E. Cicely Ridley completed in about 1992. 
! 
! EOP
!
      contains
!-----------------------------------------------------------------------
!
! BOP
! !IROUTINE: dynamo_coup 
! !INTERFACE:
      subroutine dynamo_coup
! !USES:
      use cons_module,only: dlatm,dlonm,pi_dyn,ylatm,rtd
      use fields_module,only: dynpot,phim3d 
      use magfield_module,only: im,jm,dim,djm
!
! !DESCRIPTION:
! Transform needed fields to geomagnetic coordinates
! Perform field-line integrations
! Evaluate PDE coefficients and RHS
! The PDE is divided by 1/ DT0DTS (in dyncal divided by 1/cos(theta_0)
! Sigma_(phi phi) = zigm11/ rcos0s * dt0dts
! Sigma_(lam lam) = zigm22 * rcos0s / dt0dts
! Sigma_(phi lam) = +-(zigm2-zigmc)
! Sigma_(lam phi) = -+(zigm2+zigmc)
! K_(m phi)^D     =   rim(1) * dt0dts
! K_(m lam)^D     = +-rim(2) * rcos0s
!
! !RETURN VALUE:
! !PARAMETERS: 
! !REVISION HISTORY:
! 05.02.04  <Astrid Maute> <include header> 
! 
! EOP
! 
! Local:
      integer :: i,j,jj,jjj,j0,jntl,k,n,ncc,nmaglat,nmaglon,ier
      real :: sym
      real :: array(-15:nmlon0+16,nmlat0),cs(nmlat0)
      character :: fname*10,labl*56,units*12
     
      call transf
!
! Fold southern hemisphere over on to northern  (was in transf.F version tgcm15)
! -Value at the equator is also folded therefore at the equatorial boundary
!  factor of 1/2 introduced
! -added values in array index 49-97 from equator to poles
! -SH array (1-49) has original SH values
! -reverse sign of zigmc (to be compatible with Cicely's) 
! -sign of K_(m lam)^D in southern hemisphere is reversed, therefore 
!     K_(m lam)^N - (- K_(m lam)^S)
!
! zigm11 = Sigma_(phi phi)(0)^T
! zigm22 = Sigma_(lam lam)(0)^T
! zigmc  =-Sigma_c(0)^T 
! zigm2  = Sigma_h(0)^T 
! rim(1) = K_(m phi)^D(0)^T
! rim(2) = K_(m lam)^D(0)^T
!
      do j=1,nmlath
        do i=1,nmlonp1
          zigm11(i,nmlatp1-j) = (zigm11(i,nmlatp1-j)+zigm11(i,j))
          zigmc(i,nmlatp1-j)  =  zigmc(i,nmlatp1-j)+zigmc(i,j)
          zigmc(i,nmlat+1-j)  = -zigmc(i,nmlat+1-j)
          zigm2(i,nmlat+1-j)  =  zigm2(i,nmlat+1-J)+zigm2(i,j)
          zigm22(i,nmlat+1-j) = (zigm22(i,nmlat+1-J)+zigm22(i,j))
	  
          rim(i,nmlatp1-j,1)  = (rim(i,nmlat+1-j,1)+rim(i,j,1))
          rim(i,nmlatp1-j,2)  = (rim(i,nmlat+1-j,2)+rim(i,j,2))  
        enddo ! i=1,nmlonp1
      enddo ! j=1,nmlath
!
! Calculate RHS of PDE from rim(1) and rim(2) (was in transf.F version tgcm15)
!  [( d K_(m phi)^D / d phi /(cos(theta_m)?) +
!  (d [ K_(m lam)^D * cos(lam_m)]/ d lam_m ) /cos ( lam_m) ] * R / (RCOS0S*DT0DTS)
! ~ J_(Mr)*r^2*cos(theta_m)/cos(theta_0)/DT0DTS
!
      call rhspde
!
! Set index array nc and magnetic latitude cosine array:
! nc pointes to the start of the coefficient array for each level
      nc_gip(1) = nc0
      nc_gip(2) = nc1
      nc_gip(3) = nc2
      nc_gip(4) = nc3
      nc_gip(5) = nc4
      nc_gip(6) = ncee
!
! Use pi_dyn from cons module rather than 4*atan(1) to avoid small
! differences generated by the atan in -lmass lib (-lmass was not
! used in earlier versions).
!
      do j=1,nmlat0
        cs(j) = cos(pi_dyn/2.-(nmlat0-j)*dlatm)
      enddo ! j=1,nmlat0
!
! Set up difference coefficients. 
! mixed terms: factor 4 from 5-point diff. stencil
! "NH" array index 49:97 equator to pole
! zigmc  = Sigma_(phi lam)^T(0)/( 4*d lam_0* d lon )
! zigm2  = Sigma_(phi lam)^T(0)/( 4*d lam_0* d lon )
! zigm22 = Sigma_(lam lam)^T(0)*cos(lam_0)/d lam_0^2
! zigm11 = Sigma_(phi phi)^T(0)/ cos(lam_0) / d lon^2 )
!
! "SH" array index 49:1 equator to pole
! zigmc  = -Sigma_(phi lam)^T(0)/( 4*d lam_0* d lon )
! zigm2  = -Sigma_(phi lam)^T(0)/( 4*d lam_0* d lon )
! zigm22 = Sigma_(lam lam)^T(0)*cos(lam_0)/d lam_0^2
! zigm11 = Sigma_(phi phi)^T(0)/ cos(lam_0) / d lon^2 )
! 
!
      j0 = nmlat0-nmlath
      do j=1,nmlath       !  1,49 (assuming nmlat=97)
        jj  = nmlath+j-1  ! 49,97 added values ()^T 
        jjj = nmlath-j+1  ! 49,1
!
        do i=1,nmlonp1
          zigmc(i,jj)   = (zigmc(i,jj) +zigm2(i,jj))/(4.*dlatm*dlonm)
          zigm2(i,jj)   = zigmc(i,jj)-2.*zigm2(i,jj)/(4.*dlatm*dlonm)
          zigm22(i,jj)  = zigm22(i,jj)*cs(j0+j)/dlatm**2
          zigmc(i,jjj)  = -zigmc(i,jj)
          zigm2(i,jjj)  = -zigm2(i,jj)
          zigm22(i,jjj) = zigm22(i,jj)
        enddo ! i=1,nmlonp1
        if (j /= nmlath) then
          do i = 1,nmlonp1
            zigm11(i,jj) = zigm11(i,jj)/(cs(j0+j)*dlonm**2)
            zigm11(i,jjj) = zigm11(i,jj)
          enddo
        endif
      enddo ! j=1,nmlath
!
! Set zigm11 to zero at megnetic poles to avoid floating exception 
! (values at poles are not used):
!
      do i = 1,nmlonp1
        zigm11(i,1)     = 0.0
        zigm11(i,nmlat) = 0.0
      enddo
!      
!      call addfsech_ij('ZIGM11','ZIGM11',' ',zigm11,1,nmlonp1,1,nmlat)
!      call addfsech_ij('ZIGMC' ,'ZIGMC',' ',zigmc ,1,nmlonp1,1,nmlat)
!      call addfsech_ij('ZIGM2' ,'ZIGM2',' ',zigm2 ,1,nmlonp1,1,nmlat)
!      call addfsech_ij('ZIGM22','ZIGM22',' ',zigm22,1,nmlonp1,1,nmlat)
!      call addfsech_ij('RIM1'  ,'RIM1',' ',rim(:,:,1),1,nmlonp1,1,nmlat)
!      call addfsech_ij('RIM2'  ,'RIM2',' ',rim(:,:,2),1,nmlonp1,1,nmlat)
!
! Clear array for difference stencils at all levels:
      call clearcee(cee_gip,nmlon0,nmlat0)
!
! Calculate contribution to stencils from each PDE coefficient
! isolve = 0 -> original mud version 5.
! isolve = 1 -> muh hybrid solver (only as direct solver -- slow)
! isolve = 2 -> modified mudpack solver (modified and unmodified coefficients)
!
      if (isolve==2) then
        cofum_gip(:,:,:) = 0. ! init
!
! Sigma_(phi phi)(0)/( cos(lam_0)*(d lon)^2 )
        call stencmd(zigm11(1,nmlat0),nmlon0,nmlat0,cee_gip,1)
!
! Sigma_(lam lam)(0)*cos(lam_0)/(d lam_0)^2
        call stencmd(zigm22(1,nmlat0),nmlon0,nmlat0,cee_gip,4)
!
! Sigma_(phi lam)(0)/( 4*d lam_0* d lon )
	zigmc(:,nmlat0) = -zigmc(:,nmlat0)
        call stencmd(zigmc(1,nmlat0),nmlon0,nmlat0,cee_gip,2)
	zigmc(:,nmlat0) = -zigmc(:,nmlat0)
!
! Sigma_(lam phi)(0)/( 4*d lam_0* d lon )
	zigm2(:,nmlat0) = -zigm2(:,nmlat0)
        call stencmd(zigm2(1,nmlat0),nmlon0,nmlat0,cee_gip,3)
	zigm2(:,nmlat0) = -zigm2(:,nmlat0)
!
! isolve /= 2: original or hybrid solver (only modified stencil).
      else
!
! Sigma_(phi phi)(0)/( cos(lam_0)*(d lon)^2 )
        call stencil(zigm11(1,nmlat0),nmlon0,nmlat0,cee_gip,1)
!
! Sigma_(lam lam)(0)*cos(lam_0)*/(d lam_0)^2
        call stencil(zigm22(1,nmlat0),nmlon0,nmlat0,cee_gip,4)
!
! Sigma_(phi lam)(0)/( 4*d lam_0* d lon )
	zigmc(:,nmlat0) = -zigmc(:,nmlat0)
        call stencil(zigmc(1,nmlat0),nmlon0,nmlat0,cee_gip,2)
	zigmc(:,nmlat0) = -zigmc(:,nmlat0)
!
! Sigma_(lam phi)(0)/( 4*d lam_0* d lon )
	zigm2(:,nmlat0) = -zigm2(:,nmlat0)
        call stencil(zigm2(1,nmlat0),nmlon0,nmlat0,cee_gip,3)
	zigm2(:,nmlat0) = -zigm2(:,nmlat0)
!
      endif ! isolve
!
! Insert RHS in finest stencil (formerly sub rths):
      do j = 1,nmlat0
        jj = nmlath-nmlat0+j
        do i = 1,nmlon0
          c0(i,j,10) = rhs(i,jj)
        enddo ! i = 1,nmlon0
      enddo ! j = 1,nmlat0
!
! Set boundary condition at the pole:
      call edges(c0,nmlon0,nmlat0)
      call edges(c1,nmlon1,nmlat1)
      call edges(c2,nmlon2,nmlat2)
      call edges(c3,nmlon3,nmlat3)
      call edges(c4,nmlon4,nmlat4)
      if (isolve==2) 
     |  call edges(cofum_gip,nmlon0,nmlat0)
!
! Divide stencils by cos(lam_0) (not rhs):
      call divide(c0,nmlon0,nmlat0,nmlon0,nmlat0,cs,1)
      call divide(c1,nmlon1,nmlat1,nmlon0,nmlat0,cs,1)
      call divide(c2,nmlon2,nmlat2,nmlon0,nmlat0,cs,1)
      call divide(c3,nmlon3,nmlat3,nmlon0,nmlat0,cs,1)
      call divide(c4,nmlon4,nmlat4,nmlon0,nmlat0,cs,1)
      if (isolve==2) 
     |  call divide(cofum_gip,nmlon0,nmlat0,nmlon0,nmlat0,cs,0)
!
! Set value of solution to 1. at pole:
      do i=1,nmlon0
        c0(i,nmlat0,10) = 1.
      enddo
!
! Modify stencils and RHS so that the NH high lat potential is inserted at
!  high latitude.  The SH high lat potential will be added back later.
!  pfrac = fraction of dynamo in solution in the NH. = 1 low lat, = 0 hi lat
!    cons_module: crit(1)=15, crit(2)=30 deg colats, or hi-lat > 75 deg,
!      dynamo < 60 deg, and combination between 60-75 mag lat.
! The dynamo is symmetric about the magnetic equator, but the high latitude
!  is anti-symmetric in both hemispheres.  However, since Mudpack uses the
!  NH potential pattern, then the SH potential pattern must be added
!  back into the 2-D phim before the call threed, and before it is
!  transformed to geographic coordinates.
!
      ncc = 1
      nmaglon = nmlon0
      nmaglat = nmlat0
      do n=1,5
        if (isolve==2) then
          call stenmd(nmaglon,nmaglat,cee_gip(ncc),
     |          phihm(1,nmlat0),pfrac)
        else
          call stenmod(nmaglon,nmaglat,cee_gip(ncc),
     |          phihm(1,nmlat0),pfrac)
        endif
        ncc = ncc+9*nmaglon*nmaglat
        if (n==1) ncc = ncc+nmaglon*nmaglat ! rhs is in 10th slot
        nmaglon = (nmaglon+1)/2
        nmaglat = (nmaglat+1)/2
      enddo ! n=1,5
!
      jntl = 0
!
      ier = 0 
      if(isolve==0) then
        call mud(rim,jntl,isolve,which_dynamo,ier)    ! solver in mud.F
        if(ier < 0 ) then 	! not converged
	  write(6,*) 'muh: use direct solver'
	  call muh(rim,jntl,which_dynamo) 	      ! solver in mud.F
	endif
      elseif (isolve==1) then
        call muh(rim,jntl,which_dynamo)        	     ! solver in muh2cr.F
      elseif (isolve==2) then
        call mudmod(rim,jntl,isolve,which_dynamo,ier)! solver in mudmod.F
        if(ier < 0 ) then 	! not converged
	  write(6,*) 'muh: use direct solver'
	  call muh(rim,jntl,which_dynamo) 	     ! solver in mud.F
	endif
      else
        write(6,*) 'dynamo: solver type ',isolve,' not implemented.'
        stop 'isolve'
      endif
!
! Copy output potential from rim to phim(nmlonp1,nmlat):
!  Correct the SH potential for the anti-symmetric imposed NH high lat poten
!  pfrac = fraction of dynamo in solution in the NH. = 1 low lat, = 0 hi lat
!    cons_module: crit(1)=15, crit(2)=30 deg colats, or hi-lat > 75 deg,
!      dynamo < 60 deg, and combination between 60-75 mag lat.
!  ie, need (1.-pfrac(i,nmlat0)) at phim(i,1), etc
! jn index for NH part of potential (nmlat down to ~nmlat0)
! jp index for NH pfrac (nmlat0 down to 1)
!
      do j=1,nmlat0
        jn = nmlat - j + 1
        jp = nmlat0 - j + 1
        do i=1,nmlonp1
          phim(i,j)=rim(i,j,1)+(1.-pfrac(i,jp))*(phihm(i,j)-phihm(i,jn))
!	  write(6,'(2i4,3(x,f12.5))') i,j,rim(i,j,1),pfrac(i,jp),
!     |          phihm(i,j)
        enddo ! i=1,nmlonp1
      enddo ! j=1,nmlat0
      do j=nmlat0+1,nmlat
        do i=1,nmlonp1
          phim(i,j) = rim(i,j,1)
        enddo ! i=1,nmlonp1
      enddo ! j=1,nmlat 
!      
! periodic point      
      phim(1,:) = phim(nmlonp1,:)
!      
      call addfld('PHIHM2D','2D HEELIS ELECTRIC POTENTIAL',
     |    'V',phihm,'mlon',1,nmlonp1,'mlat',1,nmlat,0)
      call addfld('PHIM2D','2D ELECTRIC POTENTIAL','VOLTS',phim,
     |    'mlon',1,nmlonp1,'mlat',1,nmlat,0)
!
! Call threed to calculate 2-d electric potential array in geomagnetic coordinates
!   from 2-d solver output phim, corrected for the SH potential
!
      call threed
!    
! Transform phim3d to geographic coordinates in dynpot (fields.F):
!   phim3d(nmlonp1,nmlat,-2:nlevp1)  ! 3d electric potential magnetic
!   dynpot(nlonp1,0:nlatp1,nlevp1),  ! 3d electric potential geographic
!
      do k=1,nlevp1
        call mag2geo(phim3d(1,1,k),dynpot(1,0,k),im(1,0),jm(1,0),
     |    dim(1,0),djm(1,0),nlonp1,nmlonp1,nlon,nlat+2,nmlon,nmlat)
      enddo ! k=1,nlevp1
!
! Periodic point:
      do k=1,nlevp1
        do j=0,nlatp1
          dynpot(nlonp1,j,k)    = dynpot(1,j,k)
        enddo ! j=0,nlatp1
      enddo ! k=1,nlevp1
!
      end subroutine dynamo_coup
!-----------------------------------------------------------------------
! BOP
! !IROUTINE: transf
! !INTERFACE:
      subroutine transf
!
! !USES:
      use cons_module,only: dt1dts,rcos0s
!
! !DESCRIPTION: 
!       - include the boundary condition at the equator
!       - transformation from lam_m to lam_0 (regular grid)
!
! !REVISION HISTORY:
! 05.02.04  <Astrid Maute> <include header> 
! 
! EOP
!
! Local:
      integer :: i,ii,k,kk,j,jj,lat,n
      real :: corfac
!
!
! External:
      real,external :: sddot ! in util.F
!
      call set_zpoten     
!
!
! From sub fieldline_integrals:
!   zigm11 is int[sig_p*d_1^2/D] ds,   i.e. Sigma_(phi phi)/abs(sin Im)
!   zigm22 is int[sig_p*d_2^2/D] ds,   i.e. Sigma_(lam lam)*abs(sin Im)
!   zigmc  is int[sig_p*d_1*d_2/D] ds, i.e. Sigma_c
!   zigm2  is int[sigma_h] ds,         i.e. Sigma_h
!
!   rim1 is int[(sigma_h-sigma_p*d_1*d_2/D)u_e1 + sigma_p*d_1^2/D u_e2] *A(h_r)*
!                B_e3 ds, i.e.  K_(m phi)^D/abs(sin Im)
!   rim2 is int[(sigma_h+sigma_p*d_1*d_2/D)u_e2 - sigma_p*d_2^2/D u_e1] *A(h_r)*
!                B_e3 ds, K_(m lam)^D ( minus in northern hemisphere
!   Change sign of RIM(2) in S. hemisphere to be compatible with transf
! At this point, rim2 is +-K_(m lam)^D
!
! am 10/04 leave this out and check what difference it makes
! but make sure that field-line integral at equator is definite
!
! Equatorial values:
! Assume that quantities primarily dependent on Pedersen conductivity
!   have field-line integrals 1/4 as large as the averages for next-higher
!   field lines; quantities primarily dependent on Hall conductivity
!   have field-line integrals 0.12 as large as the averages for next-higher
!   field lines.  Exact values chosen should not be important for potential
!   calculation, as long as they are physically reasonable and not too
!   different from adjacent values.
! add factor 1/2 for the equatorial values since both hemispheres are
! added together - two physical points but equator is only one point
!
!      j = nmlat/2+1
!      do i = 1,nmlon
!        zigm11(i,j)= .125*(zigm11(i,j-1)+ zigm11(i,j+1))
!        zigm22(i,j)= .125*(zigm22(i,j-1)+ zigm22(i,j+1))
!        zigmc(i,j) = .125*(zigmc(i,j-1) + zigmc(i,j+1))
!        zigm2(i,j) = .06 *(zigm2(i,j-1) + zigm2(i,j+1))
!        rim(i,j,1) = .06 *(rim(i,j-1,1) + rim(i,j+1,1))
!        rim(i,j,2) = .06 *(rim(i,j-1,2) + rim(i,j+1,2))
!      enddo ! i = 1,nmlon
!
! Include the boundary condition at the equator eq.(5.30) in
! Richmond (1995) Ionospheric Electrodynamics use. Mag. Apex Coord. 
!   
! Sigma_(phi phi)/|sin Im| = 0.5*Sigma_cowling/|sin Im|
!        = 0.5/ |sin Im|*[Simag_(phi phi) -
!                        Sigma_(phi lam)*Sigma_(lam phi)/Sigma_(lam lam)]
!        = 0.5/|sin Im|*[Simag_(phi phi) + 
!                       (Sigma_h-Sigma_c)*(Sigma_h + Sigma_c)/Sigma_(lam lam)]
!  K_(m phi)^D / |sin I_m| =  1/2/|sin Im|*(K_(m phi) - 
!                       Sigma_(phi lam)/Sigma_(lam lam)*K_(m lam)^D)
! factor 1/2 is taken care of above when the conductances at the
! equator are set
!
      j = nmlath      ! nmlath = (nmlat+1)/2
      do i = 1,nmlon
        zigm11(i,j) = zigm11(i,j)+ (zigm2(i,j)-zigmc(i,j))*
     |                (zigm2(i,j)+zigmc(i,j))/zigm22(i,j)
        rim(i,j,1)  = rim(i,j,1) - (zigm2(i,j)-zigmc(i,j))/
     |                zigm22(i,j)*rim(i,j,2)
        zigm11(i,j) = 0.5*zigm11(i,j)
        rim(i,j,1)  = 0.5*rim(i,j,1)
      enddo

!        
! Using notation of Richmond (1995) on right-hand side below:
! Sigma_(phi phi)/ |sin I_m|  = zigm11
! Sigma_(lam lam) * |sin I_m| = zigm22
! Sigma_(phi lam)             = +/-(zigm2-zigmc)
! Sigma_(lam phi)             = -/+(zigm2+zigmc)
! K_(m phi)^D / |sin I_m|     = rim(1)
! K_(m lam)^D                 = +/-rim(2)
!
! Transforming PDE from original apex (lam_m) to new apex grid (lam_0)
!     lam_m is irregular spaced in latitude
!     lam_0 is regular spaced in latitude (used for derivatives)
! the whole PDE is divided by d lam_m /d lam_0
! 
! sign of K_(m lam)^D in southern hemisphere remains reversed.
! for the mixed terms the transformation factors cancel out (zigmc, zigm2)
! DT1DTS : d lam_0/ d lam_m / |sin I_m|
! RCOS0S : cos(lam_0)/ cos(lam_m)
!
! corfac: |sin I_m|*d lam_m/d lam_0 * cos(lam_0)/ cos(lam_m)
! zigm11: |sin I_m|*d lam_m/d lam_0 * cos(lam_0)/ cos(lam_m)
! zigm22: 1/|sin I_m|*d lam_0/d lam_m * cos(lam_m)/ cos(lam_0)
! rim(1): |sin I_m|*d lam_m/d lam_0
! rim(2): cos(lam_m)/ cos(lam_0)        
!
      do j=2,nmlat-1
        corfac = rcos0s(j)/dt1dts(j)
        do i=1,nmlon
          zigm11(i,j) = zigm11(i,j)*corfac
          zigm22(i,j) = zigm22(i,j)/corfac
          rim(i,j,1)  = rim(i,j,1)/dt1dts(j) 
          rim(i,j,2)  = rim(i,j,2)/rcos0s(j) 
        enddo ! i,nmlon
      enddo ! j=2,nmlat-1
!
! Periodic points:
      do j=1,nmlat
        zigm11(nmlonp1,j) = zigm11(1,j)
        zigmc (nmlonp1,j) = zigmc (1,j)
        zigm2 (nmlonp1,j) = zigm2 (1,j)
        zigm22(nmlonp1,j) = zigm22(1,j)
        rim(nmlonp1,j,:)  = rim(1,j,:)
      enddo ! j=1,nmlat
!
! zigm11 = Sigma_(phi phi)(0)= Sigma_(phi phi)(m) * cos lam_0 /cos lam_m * d lam_m /d lam_0
! zigm22 = Sigma_(lam lam)(0)= Sigma_(lam lam)(m) * cos lam_m /cos lam_0 * d lam_0 /d lam_m
! +-(zigm2-zigmc)= Sigma_(phi lam)(0) = Sigma_(phi lam)(m)
! -+(zigm2+zigmc)= Sigma_(lam phi)(0) = Sigma_(lam phi)(m)
! rim(1) = K_(m phi)^D(0)    = K_(m phi)^D(m) * d lam_m /d lam_0   
! rim(2) = +/-K_(m lam)^D(0) = +/-K_(m lam)^D(m) * cos lam_m /cos lam_0
!  
! Compute polar values for the conductances, 4th order interpolation:
! 
      zigm11(1,    1) = (4.*sddot(nmlon,unitvm,zigm11(1,      2))-
     1  sddot(nmlon,unitvm,zigm11(1,      3)))/(3.*float(nmlon))
      zigm11(1,nmlat) = (4.*sddot(nmlon,unitvm,zigm11(1,nmlat-1))-
     1  sddot(nmlon,unitvm,zigm11(1,nmlat-2)))/(3.*float(nmlon))
      zigmc(1,    1) = (4.*sddot(nmlon,unitvm,zigmc(1,      2))-
     1  sddot(nmlon,unitvm,zigmc(1,      3)))/(3.*float(nmlon))
      zigmc(1,nmlat) = (4.*sddot(nmlon,unitvm,zigmc(1,nmlat-1))-
     1  sddot(nmlon,unitvm,zigmc(1,nmlat-2)))/(3.*float(nmlon))
      zigm2(1,    1) = (4.*sddot(nmlon,unitvm,zigm2(1,      2))-
     1  sddot(nmlon,unitvm,zigm2(1,      3)))/(3.*float(nmlon))
      zigm2(1,nmlat) = (4.*sddot(nmlon,unitvm,zigm2(1,nmlat-1))-
     1  sddot(nmlon,unitvm,zigm2(1,nmlat-2)))/(3.*float(nmlon))
      zigm22(1,    1) = (4.*sddot(nmlon,unitvm,zigm22(1,      2))-
     1  sddot(nmlon,unitvm,zigm22(1,      3)))/(3.*float(nmlon))
      zigm22(1,nmlat) = (4.*sddot(nmlon,unitvm,zigm22(1,nmlat-1))-
     1  sddot(nmlon,unitvm,zigm22(1,nmlat-2)))/(3.*float(nmlon))
! 
! Extend over longitude                                        
      do i = 2,nmlon
        zigm11(i,    1)  = zigm11(1,    1)
        zigm11(i,nmlat)  = zigm11(1,nmlat)
        zigmc(i,     1)  = zigmc(1,     1)
        zigmc(i, nmlat)  = zigmc(1, nmlat)
        zigm2(i,     1)  = zigm2(1,     1)
        zigm2(i, nmlat)  = zigm2(1, nmlat)
        zigm22(i,     1) = zigm22(1,    1)
        zigm22(i, nmlat) = zigm22(1,nmlat)
      enddo ! i = 2,nmlon
!
! RHS vector (I_1,I_2): average over poles:
      do i = 1,nmlon
        rim(i,1,1) = .5*(rim(i,2,1)-rim(1+mod(i-1+nmlon/2,nmlon),2,1))
        rim(i,nmlat,1) = .5*(rim(i,nmlat-1,1)-
     |    rim(1+mod(i-1+nmlon/2,nmlon),nmlat-1,1))
        rim(i,1,2) = .5*(rim(i,2,2)-rim(1+mod(i-1+nmlon/2,nmlon),2,2))
        rim(i,nmlat,2) = .5*(rim(i,nmlat-1,2)-
     |    rim(1+mod(i-1+nmlon/2,nmlon),nmlat-1,2))
      enddo ! i = 1,nmlon
! 
! Periodic points:
      do j=1,nmlat
        zigm11(nmlonp1,j) = zigm11(1,j)
        zigmc (nmlonp1,j) = zigmc (1,j)
        zigm2 (nmlonp1,j) = zigm2 (1,j)
        zigm22(nmlonp1,j) = zigm22(1,j)
        rim(nmlonp1,j,:)  = rim(1,j,:)
      enddo ! j=1,nmlat
!
! Save to secondary history:
!     real,dimension(nmlonp1,nmlat) ::
!    |  zigm11,  ! sigma11*cos(theta0)
!    |  zigmc,   ! sigmac
!    |  zigm2,   ! sigma2
!    |  zigm22   ! sigma22/cos(theta0)
!
!     do j=1,nmlat
!       do i=1,nmlonp1
!         zigm11_plt(i,:) = zigm11(i,j)
!         zigmc_plt (i,:) = zigmc (i,j)
!         zigm2_plt (i,:) = zigm2 (i,j)
!         zigm22_plt(i,:) = zigm22(i,j)
!         rim1_plt  (i,:) = rim   (i,j,1)
!         rim2_plt  (i,:) = rim   (i,j,2)
!       enddo ! i=1,nmlonp1
! 
! Folding south on to northern hemisphere and calculation of RHS has 
!   been moved to subroutine dynamo.
!
      end subroutine transf
!-----------------------------------------------------------------------
! BOP
! !IROUTINE: rhspde
! !INTERFACE:
      subroutine rhspde
! !USES:
      use cons_module,only: pi_dyn,dlatm,dlonm,r0
! !DESCRIPTION:
!  differentiate RHS 
!  R_0 * d lam_m/d lam_0 * 1/cos lam_0 [ d K_(m phi)^DT(0) / d phi  +
!  (d [ K_(m lam)^DT(0) * cos(lam_0)]/ d lam_0  ] 
!
! !REVISION HISTORY:
! 05.02.16  <Astrid Maute> <include header> 
! 
! EOP
!
! Local:
      integer :: i,j,jj
      real :: pi
      real,dimension(nmlat) :: tint1,tint2,tint3
      real,dimension(-1:nmlonp1+1) :: tint33
!
! External:
      real,external :: sddot ! in util.F
!
      pi = pi_dyn
!
! cos lam_0:
      do j=1,nmlat
        tint1(j) = cos(-pi/2.+(j-1)*dlatm)
      enddo ! j=1,nmlat
!
! Perform differentiation (1/ cos lam0 ) * d K_(m phi)^DT(0)/ d phi
! central differencing
! rhs(i,j) = 1/ cos lam0(j) *  [K_(m phi)^DT(0)(i+1,j)-K_(m phi)^DT(0)(i-1,j)]/ (2 d lon)
! rim(1) =  K_(m phi)^DT(0)
! copy rim(1) into tint33 and set up wrap around points
!
      do j=2,nmlath-1   !  2,48                            
        jj = j+nmlath-1 ! 49,96
        do i=1,nmlon
          tint33(i) = rim(i,jj,1)         ! tint33(1:nmlon)
        enddo ! i=1,nmlon
        do i=1,2                          ! wrap around points
          tint33(i-2) = tint33(i-2+nmlon) ! -1:0 <= nmlon-1:nmlon
          tint33(i+nmlon) = tint33(i)     ! nmlon+1:nmlon+2 <= 1,2
        enddo ! i=1,2
        do i=1,nmlon                                         
          rhs(i,j) = 1./(dlonm*tint1(nmlath+j-1))*
     |      .5*(tint33(i+1)-tint33(i-1))  ! tint33 2:nmlon+1 and 0:nmlon-1
        enddo ! i=1,nmlon
      enddo ! j=2,nmlath-1
!
! Perform differentiation (1/ cos lam0 ) * d [K_(m lam)^DT(0) cos lam_0]/ d lam_0:
! central differencing
! rhs(i,j) = rhs(i,j) + 1/ cos lam0(j) *  
!  [K_(m lam)^DT(0)(i,j+1)cos lam0(j+1)-K_(m lam)^DT(0)(i,j-1)cos lam0(j-1)]/ (2 d lat)
!
      do j=nmlath+1,nmlat-1 ! 50,96                          
        jj = j-nmlath+1     !  2,48
        do i=1,nmlon
          rhs(i,jj) = rhs(i,jj)+1./(dlatm*tint1(j))*
     |      .5*(rim(i,j+1,2)*tint1(j+1)-rim(i,j-1,2)*tint1(j-1))
        enddo ! i=1,nmlon
      enddo ! j=nmlath+1,nmlat-1
!
! polar value:
      rhs(1,nmlath) = -2./float(nmlon)*
     |  sddot(nmlon,unitvm,rim(1,nmlat-1,2))/tint1(nmlat-1)
!
! equatorial value:
! [K_(m phi)^DT(0)(i+1,j)-K_(m phi)^DT(0)(i-1,j)]/ (2 d lon) +
! [K_(m lam)^DT(0)(i,j+1/2)cos lam0(j+1/2)-0]/ ( d lat/2)
!   with   -  K_(m lam)^DT(0)(i,j_eq) = 0
!          -  cos lam0(j_eq) =1
!
      i = 1
      rhs(i,1) = 0.5/dlonm*(rim(i+1,nmlath,1)-rim(nmlon,nmlath,1))
      rhs(i,1) = rhs(i,1)+1./dlatm*(tint1(nmlath)*rim(i,nmlath,2)+
     |                            tint1(nmlath+1)*rim(i,nmlath+1,2))
      do i = 2,nmlon-1
        rhs(i,1) = 0.5/dlonm*(rim(i+1,nmlath,1)-rim(i-1,nmlath,1))
        rhs(i,1) = rhs(i,1)+1./dlatm*(tint1(nmlath)*rim(i,nmlath,2)+
     |                              tint1(nmlath+1)*rim(i,nmlath+1,2))
      enddo ! i = 2,nmlon-1
      i = nmlon
      rhs(i,1) = 0.5/dlonm*(rim(1,nmlath,1)-rim(i-1,nmlath,1))
      rhs(i,1) = rhs(i,1)+1./dlatm*(tint1(nmlath)*rim(i,nmlath,2)+
     |                            tint1(nmlath+1)*rim(i,nmlath+1,2))
!
! Extend over longitude:
      do i=2,nmlon
        rhs(i,nmlath) = rhs(1,nmlath)
      enddo ! i=2,nmlon
!
! Periodic points:
      do j=1,nmlath
        rhs(nmlonp1,j) = rhs(1,j)
      enddo ! j=1,nmlath
!
! multiply by earth radius + h_0 in meter  = R0*1.E-2 
!  R_0 * d lam_m/d lam_0 * 1/cos lam_0 [ d K_(m phi)^DT(0) / d phi  +
!  (d [ K_(m lam)^DT(0) * cos(lam_0)]/ d lam_0  ] 
!
      do j=1,nmlath
        do i=1,nmlonp1
          rhs(i,j) = rhs(i,j)*r0*1.e-2
        enddo ! i=1,nmlonp1
      enddo ! j=1,nmlath
!
! Save rhs to secondary histories (redundant in vertical and
!  zero in north hem):
!     real,dimension(nmlonp1,nmlath)  :: rhs ! right-hand side 
!     real,dimension(nmlonp1,-2:nlev) :: rhs_plt ! diag
!
!     rhs_plt = 0.
!     do j=1,nmlat
!       if (j <= nmlath) then
!         do i=1,nmlonp1
!           rhs_plt(i,:) = rhs(i,j)
!         enddo
!       endif
!       call addfsech_ik('RHS',' ',' ',rhs_plt,
!    |    1,nmlonp1,nmlev,nmlev-1,j)
!     enddo ! j=1,nmlat
!
      end subroutine rhspde
!-----------------------------------------------------------------------
 ! BOP
! !IROUTINE: clearce
! !INTERFACE:
      subroutine clearcee(cee,nlon0,nlat0)
!
! !DESCRIPTION:
! Zero C arrays for stencil coefficients.
! Cee will contain:
!   c0(nmlon0,nmlat0,10), c1(nmlon1,nmlat1,9), c2(nmlon2,nmlat2,9),
!   c3(nmlon3,nmlat3,9),  c4(nmlon4,nmlat4,9)
!
! !PARAMETERS: 
      integer,intent(in) :: nlon0,nlat0
! !RETURN VALUE:
      real,intent(out) :: cee(*)
!
! !REVISION HISTORY:
! 05.03.8  <Astrid Maute> <include header> 
! 
! EOP
!
! Local:
      integer :: nlon,nlat,n,m,i
!
! Compute total size of cee
      nlon = nlon0
      nlat = nlat0
      n = 0
      do m=1,5 ! 5 resolution levels
        n = n+nlon*nlat
        nlon = (nlon+1)/2
        nlat = (nlat+1)/2
      enddo ! m=1,5 ! 5 resolution levels
      n = 9*n+nlon0*nlat0
!
! Clear cee:
      do i=1,n
        cee(i) = 0.
      enddo
      end subroutine clearcee
!-----------------------------------------------------------------------
 ! BOP
! !IROUTINE: stencil
! !INTERFACE:
      subroutine stencil(zigm,nlon0,nlat0,cee,ncoef)
!
! !DESCRIPTION:
! Calculate contribution fo 3 by 3 stencil from coefficient zigm
! at each grid point and level.
!
! !PARAMETERS: 
      integer,intent(in) :: 
     |  nlon0, ! longitude dimension of finest grid level
     |  nlat0, ! latitude dimension of finest grid level
     |  ncoef   ! integer identifier of coefficient
      real,intent(in) :: 
     |  zigm(nlon0,nlat0)  ! coefficients (nlon0+1/2,(nlat0+1)/2) 
!     
! !RETURN VALUE:
      real,intent(inout) :: 
     |  cee(*)  ! output stencil array consisting of c0,c1,c2,c3,c4
!
! Local:
      integer :: nc,nlon,nlat,n
!
! Perform half-way interpolation and extend zigm in array:
!
      call htrpex(zigm,nlon0,nlat0)
!
! Calculate contribution to stencil for each grid point and level:
      nc = 1
      nlon = nlon0
      nlat = nlat0
      do n=1,5 ! 5 levels of resolution
        call cnm(nlon0,nlat0,nlon,nlat,cee(nc),ncoef)
        nc = nc+9*nlon*nlat
        if (n==1) nc = nc+nlon*nlat
        nlon = (nlon+1)/2
        nlat = (nlat+1)/2
      enddo ! n=1,5
      end subroutine stencil
!-----------------------------------------------------------------------
! BOP
! !IROUTINE: stencmd
! !INTERFACE:
      subroutine stencmd(zigm,nlon0,nlat0,cee,ncoef)
!
! !DESCRIPTION: 
! Calculate contribution fo 3 by 3 stencil from coefficient zigm
! at each grid point and level.
!
! !RETURN VALUE:
      integer,intent(in) :: 
     |  nlon0, ! longitude dimension of finest grid level
     |  nlat0, ! latitude dimension of finest grid level
     |  ncoef  ! integer identifier of coefficient
      real,intent(in) :: 
     |  zigm(nlon0,nlat0)  ! coefficients (nlon0+1/2,(nlat0+1)/2) 
!     
! !RETURN VALUE:
      real,intent(inout) :: 
     |  cee(*)  ! output stencil array consisting of c0,c1,c2,c3,c4
!       
! !REVISION HISTORY:
! 18.02.05  <Astrid Maute> <include header> 
! 
! EOP
! Local:
      integer :: nc,nlon,nlat,n
!
! Perform half-way interpolation and extend zigm in array:
!
      call htrpex(zigm,nlon0,nlat0)
!
! Calculate contribution to stencil for each grid point and level:
!
      nc = 1
      nlon = nlon0
      nlat = nlat0
!
! Calculate modified and unmodified stencil on finest grid
!
      call cnmmod(nlon0,nlat0,nlon,nlat,cee(nc),ncoef)
!
! Stencils on other grid levels remain the same.
      nc = nc+10*nlon*nlat 
      nlon = (nlon+1)/2
      nlat = (nlat+1)/2
!
      do n=2,5
        call cnm(nlon0,nlat0,nlon,nlat,cee(nc),ncoef)
        nc = nc+9*nlon*nlat
        if (n==1) nc = nc+nlon*nlat
        nlon = (nlon+1)/2
        nlat = (nlat+1)/2
      enddo ! n=1,5
      end subroutine stencmd
!-----------------------------------------------------------------------
! BOP
! !IROUTINE: htrpex
! !INTERFACE:
      subroutine htrpex(coeff,nmlon0,nmlat0)
!
! !DESCRIPTION: 
! copy coefficients into array and
! extend array over 16 grid points for the wrap around points
! on the 5 different grid levels. 
! Result returned in array(1:nmlath) - equator to pole.
!
! !ARGUMENTS:
      integer,intent(in) :: nmlon0,nmlat0        ! nmlon, nmlath
      real,intent(in) :: coeff(nmlon0,nmlat0)    ! Sigma(SH) pole to equator
!        
! !REVISION HISTORY:
! 18.02.05  <Astrid Maute> <include header> 
! 
! EOP
!
! Local:
      integer :: i,j,jj
!
! Copy coeff into positions in array:
! MODIFY: NO SYM. USE NH VALUES
!
      do j=1,nmlat0      ! 1:nmlath
        do i=1,nmlon0
          wkarray(i,j) = coeff(i,j)
        enddo ! i=1,nmlon0
      enddo ! j=1,nmlat0
!
! Extend over 32 grid spaces to allow for wrap around points for
! a total of 5 grid levels:
!
      do i=1,16
        do j=1,nmlat0
          wkarray(1-i,j)      = wkarray(nmlon0-i,j) 
          wkarray(nmlon0+i,j) = wkarray(1+i,j)
        enddo ! j=1,nmlat0
      enddo ! i=1,16
!
      end subroutine htrpex
!-----------------------------------------------------------------------
! BOP
! !IROUTINE: cnm
! !INTERFACE:
      subroutine cnm(nlon0,nlat0,nlon,nlat,c,ncoef)
!
! !DESCRIPTION: 
! Compute contribution to stencil from zigm(ncoef) on grid nlon by nlat,
! Finest grid is nlon0 by nlat0.
! uses upwinding method if diagonal dominance is not preserved
!
! ncoef: integer id of Sigma coefficient:
! ncoef = 1 for zigm11 Sigma_(phi phi)
! ncoef = 2 for zigmc  Sigma_(phi lam)
! ncoef = 3 for zigm2  Sigma_(lam phi)
! ncoef = 4 for zigm22 Sigma_(lam lam)
!
! !ARGUMENTS:
      integer,intent(in) :: 
     |  nlon0,nlat0, ! finest grid dimensions
     |  nlon,nlat    ! output grid dimensions
      integer,intent(in) :: ncoef

! !RETURN VALUE:
      real,intent(inout) :: 
     |  c(nlon,nlat,*)    ! output array for grid point stencils at
                          ! resolution nlon x nlat
!
! !REVISION HISTORY:
! 18.02.05  <Astrid Maute> <include header> 
! 
! EOP
! Local:
      integer :: i,j,nint,i0,j0
      real,parameter :: pi=3.141592654
      real :: wk(nlon0,3)
!
! Compute separation of grid points of resolution nlon x nlat within
! grid of resolution nlon0,nlat0. Evaluate dlon and dlat, grid spacing 
! of nlon x nlat.
!
      nint = (nlon0-1)/(nlon-1)
!
! Scan wkarray nlon x nlat calculating and adding contributions to stencil
! from zigm(ncoef)
      i0 = 1-nint
      j0 = 1-nint
!
! zigm11 Sigma_(phi phi)
      if (ncoef==1) then
        do j = 1,nlat-1
          do i = 1,nlon
            c(i,j,1) = c(i,j,1)+.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+(i+1)*nint,j0+j*nint))
            c(i,j,5) = c(i,j,5)+.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+(i-1)*nint,j0+j*nint))
            c(i,j,9) = c(i,j,9)-.5*(wkarray(i0+(i+1)*nint,j0+j*nint)+
     |        2.*wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+(i-1)*nint,j0+j*nint))
          enddo ! i = 1,nlon
        enddo ! j = 2,nlat-1
!
! zigmc Sigma_(phi lam)
      elseif (ncoef==2) then
        do j = 2,nlat-1
          do i = 1,nlon
            c(i,j,2) = c(i,j,2)+.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+(i+1)*nint,j0+j*nint))
            c(i,j,4) = c(i,j,4)-.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+(i-1)*nint,j0+j*nint))
            c(i,j,6) = c(i,j,6)+.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+(i-1)*nint,j0+j*nint))
            c(i,j,8) = c(i,j,8)-.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+(i+1)*nint,j0+j*nint))
            wk(i,1) = .5*(wkarray(i0+(i+1)*nint,j0+j*nint)-
     |        wkarray(i0+(i-1)*nint,j0+j*nint))
            wk(i,2) = (c(i,j,3)+wk(i,1))*(c(i,j,7)-wk(i,1))
            wk(i,3) = sign(wk(i,1),c(i,j,3)+c(i,j,7))
            if (wk(i,2) >= 0.) wk(i,3) = 0.
            c(i,j,3) = c(i,j,3)+wk(i,1)+wk(i,3)
            c(i,j,7) = c(i,j,7)-wk(i,1)+wk(i,3)
            c(i,j,9) = c(i,j,9)-2.*wk(i,3)
          enddo ! i = 1,nlon
        enddo ! j = 2,nlat-1
!
! zigm2  Sigma_(lam phi)
      elseif (ncoef==3) then
        do j = 2,nlat-1
          do i = 1,nlon
            c(i,j,2) = c(i,j,2)+.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+i*nint,j0+(j+1)*nint))
            c(i,j,4) = c(i,j,4)-.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+i*nint,j0+(j+1)*nint))
            c(i,j,6) = c(i,j,6)+.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+i*nint,j0+(j-1)*nint))
            c(i,j,8) = c(i,j,8)-.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+i*nint,j0+(j-1)*nint))
            wk(i,1) = .5*(wkarray(i0+i*nint,j0+(j+1)*nint)-
     |        wkarray(i0+i*nint,j0+(j-1)*nint))
            wk(i,2) = (c(i,j,1)+wk(i,1))*(c(i,j,5)-wk(i,1))
            wk(i,3) = sign(wk(i,1),c(i,j,1)+c(i,j,5))
            if (wk(i,2) >= 0.) wk(i,3) = 0.
            c(i,j,1) = c(i,j,1)+wk(i,1)+wk(i,3)
            c(i,j,5) = c(i,j,5)-wk(i,1)+wk(i,3)
            c(i,j,9) = c(i,j,9)-2.*wk(i,3)
          enddo ! i = 1,nlon
        enddo ! j = 2,nlat-1
!
! Low latitude boundary condition:
        j = 1
        do i=1,nlon
          c(i,j,2) = c(i,j,2)+.5*(wkarray(i0+i*nint,j0+j*nint)+
     |      wkarray(i0+i*nint,j0+(j+1)*nint))
          c(i,j,4) = c(i,j,4)-.5*(wkarray(i0+i*nint,j0+j*nint)+
     |      wkarray(i0+i*nint,j0+(j+1)*nint))
          wk(i,1) = .5*(wkarray(i0+i*nint,j0+j*nint)+
     |      wkarray(i0+i*nint,j0+(j+1)*nint))
          wk(i,2) = (c(i,j,1)+wk(i,1))*(c(i,j,5)-wk(i,1))
          wk(i,3) = sign(wk(i,1),c(i,j,1)+c(i,j,5))
          if (wk(i,2) >= 0.) wk(i,3) = 0.
          c(i,j,1) = c(i,j,1)+wk(i,1)+wk(i,3)
          c(i,j,5) = c(i,j,5)-wk(i,1)+wk(i,3)
          c(i,j,9) = c(i,j,9)-2.*wk(i,3)
        enddo ! i=1,nlon
!
! zigm22 Sigma_(lam lam)
      elseif (ncoef==4) then
        do j = 2,nlat-1
          do i = 1,nlon
            c(i,j,3) = c(i,j,3)+.5*(wkarray(i0+i*nint,j0+j*nint)
     |        +wkarray(i0+i*nint,j0+(j+1)*nint))
            c(i,j,7) = c(i,j,7)+.5*(wkarray(i0+i*nint,j0+j*nint)
     |        +wkarray(i0+i*nint,j0+(j-1)*nint))
            c(i,j,9) = c(i,j,9)-.5*(wkarray(i0+i*nint,j0+(j-1)*nint)
     |        +2.*wkarray(i0+i*nint,j0+j*nint)
     |        +wkarray(i0+i*nint,j0+(j+1)*nint))
          enddo ! i = 1,nlon
        enddo ! j = 2,nlat-1
!
! Low latitude boundary condition:
        j = 1
        do i=1,nlon
          c(i,j,3) = c(i,j,3)+.5*(wkarray(i0+i*nint,j0+j*nint)
     |      +wkarray(i0+i*nint,j0+(j+1)*nint))
          c(i,j,9) = c(i,j,9)-.5*(wkarray(i0+i*nint,j0+j*nint)+
     |                            wkarray(i0+i*nint,j0+(j+1)*nint))
        enddo ! i=1,nlon
      endif ! ncoef
!      
      end subroutine cnm
!-----------------------------------------------------------------------
! BOP
! !IROUTINE: cnmmod
! !INTERFACE:
      subroutine cnmmod(nlon0,nlat0,nlon,nlat,c,ncoef)
!
! !DESCRIPTION: 
! Compute contribution to stencil from zigm(ncoef) on finest grid nlon0 by nlat0.
! calculates the stencil with (c) and without (cofum)
! the use of upwinding method if diagonal dominance is not preserved
! output: c     ( upwinding)
!         cofum (no upwinding)
!
! ncoef: integer id of Sigma coefficient:
! ncoef = 1 for zigm11 Sigma_(phi phi)
! ncoef = 2 for zigmc  Sigma_(phi lam)
! ncoef = 3 for zigm2  Sigma_(lam phi)
! ncoef = 4 for zigm22 Sigma_(lam lam)
!
! !ARGUMENTS:
      integer,intent(in) :: 
     |  nlon0,nlat0, ! finest grid dimensions
     |  nlon,nlat    ! output grid dimensions
      integer,intent(in) :: ncoef
! !RETURN VALUE:
      real,intent(inout) :: 
     |  c(nlon,nlat,*)    ! output array for grid point stencils at
                          ! resolution nlon x nlat
!        
! !REVISION HISTORY:
! 18.02.05  <Astrid Maute> <include header> 
! 
! EOP
!
! Local:
      integer :: i,j,nint,i0,j0
      real,parameter :: pi=3.141592654
      real :: wk(nlon0,3)
!
! Compute separation of grid points of resolution nlon x nlat within
! grid of resolution nlon0,nlat0. Evaluate dlon and dlat, grid spacing 
! of nlon x nlat.
!
      nint = (nlon0-1)/(nlon-1)
!
! Scan wkarray nlon x nlat calculating and adding contributions to stencil
! from zigm(ncoef)
      i0 = 1-nint
      j0 = 1-nint
!
! zigm11: Sigma_(phi phi)
      if (ncoef==1) then
        do j = 1,nlat-1
          do i = 1,nlon
            c(i,j,1) = c(i,j,1)+.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+(i+1)*nint,j0+j*nint))
            c(i,j,5) = c(i,j,5)+.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+(i-1)*nint,j0+j*nint))
            c(i,j,9) = c(i,j,9)-.5*(wkarray(i0+(i+1)*nint,j0+j*nint)+
     |        2.*wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+(i-1)*nint,j0+j*nint))
!
! Unmodified:
            cofum_gip(i,j,1)=cofum_gip(i,j,1)+.5*(wkarray(i0+i*nint,j0
     |        +j*nint)+wkarray(i0+(i+1)*nint,j0+j*nint))
            cofum_gip(i,j,5)=cofum_gip(i,j,5)+.5*(wkarray(i0+i*nint,j0
     |        +j*nint)+wkarray(i0+(i-1)*nint,j0+j*nint))
            cofum_gip(i,j,9) = cofum_gip(i,j,9)-
     |        .5*(wkarray(i0+(i+1)*nint,j0+j*nint)+
     |         2.*wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+(i-1)*nint,j0+j*nint))
          enddo ! i = 1,nlon
        enddo ! j = 2,nlat-1
!
! zigmc Sigma_(phi lam)
      elseif (ncoef==2) then
        do j = 2,nlat-1
          do i = 1,nlon
            c(i,j,2) = c(i,j,2)+.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+(i+1)*nint,j0+j*nint))
            c(i,j,4) = c(i,j,4)-.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+(i-1)*nint,j0+j*nint))
            c(i,j,6) = c(i,j,6)+.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+(i-1)*nint,j0+j*nint))
            c(i,j,8) = c(i,j,8)-.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+(i+1)*nint,j0+j*nint))
            wk(i,1) = .5*(wkarray(i0+(i+1)*nint,j0+j*nint)-
     |        wkarray(i0+(i-1)*nint,j0+j*nint))
!
! Unmodified:
            cofum_gip(i,j,2) = c(i,j,2)
            cofum_gip(i,j,4) = c(i,j,4)
            cofum_gip(i,j,6) = c(i,j,6)
            cofum_gip(i,j,8) = c(i,j,8)
            cofum_gip(i,j,3) = cofum_gip(i,j,3)+wk(i,1)
            cofum_gip(i,j,7) = cofum_gip(i,j,7)-wk(i,1)
! test for upwinding
            wk(i,2) = (c(i,j,3)+wk(i,1))*(c(i,j,7)-wk(i,1))
            wk(i,3) = sign(wk(i,1),c(i,j,3)+c(i,j,7))
            if (wk(i,2) >= 0.) wk(i,3) = 0.
            c(i,j,3) = c(i,j,3)+wk(i,1)+wk(i,3)
            c(i,j,7) = c(i,j,7)-wk(i,1)+wk(i,3)
            c(i,j,9) = c(i,j,9)-2.*wk(i,3)
          enddo ! i = 1,nlon
        enddo ! j = 2,nlat-1
!
! zigm2 Sigma_(lam phi)
      elseif (ncoef==3) then
        do j = 2,nlat-1
          do i = 1,nlon
            c(i,j,2) = c(i,j,2)+.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+i*nint,j0+(j+1)*nint))
            c(i,j,4) = c(i,j,4)-.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+i*nint,j0+(j+1)*nint))
            c(i,j,6) = c(i,j,6)+.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+i*nint,j0+(j-1)*nint))
            c(i,j,8) = c(i,j,8)-.5*(wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+i*nint,j0+(j-1)*nint))
            wk(i,1) = .5*(wkarray(i0+i*nint,j0+(j+1)*nint)-
     |        wkarray(i0+i*nint,j0+(j-1)*nint))
!
! Unmodified:
            cofum_gip(i,j,2) = c(i,j,2)
            cofum_gip(i,j,4) = c(i,j,4)
            cofum_gip(i,j,6) = c(i,j,6)
            cofum_gip(i,j,8) = c(i,j,8)
            cofum_gip(i,j,1) = cofum_gip(i,j,1)+wk(i,1)
            cofum_gip(i,j,5) = cofum_gip(i,j,5)-wk(i,1)
! test for upwinding
            wk(i,2) = (c(i,j,1)+wk(i,1))*(c(i,j,5)-wk(i,1))
            wk(i,3) = sign(wk(i,1),c(i,j,1)+c(i,j,5))
            if (wk(i,2) >= 0.) wk(i,3) = 0.
            c(i,j,1) = c(i,j,1)+wk(i,1)+wk(i,3)
            c(i,j,5) = c(i,j,5)-wk(i,1)+wk(i,3)
            c(i,j,9) = c(i,j,9)-2.*wk(i,3)
          enddo ! i = 1,nlon
        enddo ! j = 2,nlat-1
!
! Low latitude boundary condition:
        j = 1
        do i=1,nlon
          c(i,j,2) = c(i,j,2)+.5*(wkarray(i0+i*nint,j0+j*nint)+
     |      wkarray(i0+i*nint,j0+(j+1)*nint))
          c(i,j,4) = c(i,j,4)-.5*(wkarray(i0+i*nint,j0+j*nint)+
     |      wkarray(i0+i*nint,j0+(j+1)*nint))
          wk(i,1) = .5*(wkarray(i0+i*nint,j0+j*nint)+
     |      wkarray(i0+i*nint,j0+(j+1)*nint))

          cofum_gip(i,j,2) = c(i,j,2)
          cofum_gip(i,j,4) = c(i,j,4)
          cofum_gip(i,j,1) = cofum_gip(i,j,1)+wk(i,1)
          cofum_gip(i,j,5) = cofum_gip(i,j,5)-wk(i,1)

          wk(i,2) = (c(i,j,1)+wk(i,1))*(c(i,j,5)-wk(i,1))
          wk(i,3) = sign(wk(i,1),c(i,j,1)+c(i,j,5))
          if (wk(i,2) >= 0.) wk(i,3) = 0.
          c(i,j,1) = c(i,j,1)+wk(i,1)+wk(i,3)
          c(i,j,5) = c(i,j,5)-wk(i,1)+wk(i,3)
          c(i,j,9) = c(i,j,9)-2.*wk(i,3)
        enddo ! i=1,nlon
!
! zigm22: Sigma_(lam lam)
      elseif (ncoef==4) then
        do j = 2,nlat-1
          do i = 1,nlon
            c(i,j,3) = c(i,j,3)+.5*(wkarray(i0+i*nint,j0+j*nint)
     |        +wkarray(i0+i*nint,j0+(j+1)*nint))
            c(i,j,7) = c(i,j,7)+.5*(wkarray(i0+i*nint,j0+j*nint)
     |        +wkarray(i0+i*nint,j0+(j-1)*nint))
            c(i,j,9) = c(i,j,9)-.5*(wkarray(i0+i*nint,j0+(j-1)*nint)
     |        +2.*wkarray(i0+i*nint,j0+j*nint)
     |        +wkarray(i0+i*nint,j0+(j+1)*nint))
!
! Unmodified:
            cofum_gip(i,j,3) = cofum_gip(i,j,3)+.5*(wkarray(i0+i*nint,j0
     |        +j*nint)+wkarray(i0+i*nint,j0+(j+1)*nint))
            cofum_gip(i,j,7) = cofum_gip(i,j,7)+.5*(wkarray(i0+i*nint,j0
     |        +j*nint)+wkarray(i0+i*nint,j0+(j-1)*nint))
            cofum_gip(i,j,9) = cofum_gip(i,j,9)-.5*(wkarray(i0+i*nint,j0
     |       +(j-1)* nint)+2.*wkarray(i0+i*nint,j0+j*nint)+
     |        wkarray(i0+i*nint,j0+(j+1)*nint))

          enddo ! i = 1,nlon
        enddo ! j = 2,nlat-1
!
! Low latitude boundary condition:
        j = 1
        do i=1,nlon
          c(i,j,3) = c(i,j,3)+.5*(wkarray(i0+i*nint,j0+j*nint)
     |      +wkarray(i0+i*nint,j0+(j+1)*nint))
          c(i,j,9) = c(i,j,9)-.5*(wkarray(i0+i*nint,j0+j*nint)+
     |                            wkarray(i0+i*nint,j0+(j+1)*nint))
          cofum_gip(i,j,3) = cofum_gip(i,j,3)+.5*(wkarray(i0+i*nint,j0+
     |                         j*nint)+wkarray(i0+i*nint,j0+(j+1)*nint))
          cofum_gip(i,j,9) = cofum_gip(i,j,9)-.5*(wkarray(i0+i*nint,j0+j
     |                         *nint)+wkarray(i0+i*nint,j0+(j+1)*nint))

        enddo ! i=1,nlon
      endif ! ncoef
!
      end subroutine cnmmod
!-----------------------------------------------------------------------
! BOP
! !IROUTINE: divide
! !INTERFACE:
      subroutine divide(c,nlon,nlat,nlon0,nlat0,cs,igrid)
!
! !USES:
! !DESCRIPTION: 
! Divide stencil C by cos(theta(i,j))
!
! !ARGUMENTS:
      integer,intent(in) :: nlon,nlat,   ! dim. of finest grid
     |    nlon0,nlat0,                   ! dim. of actual grid
     |    igrid                          ! flag for divide by cos lam_0
      real,intent(in) :: cs(*)           ! cos lam_0
! !RETURN VALUE:
      real,intent(out) :: c(nlon,nlat,*) ! coefficient array for grid level
!        
! !REVISION HISTORY:
! 21.02.05  <Astrid Maute> <include header> 
!                          <remove 1/cos of RHS at the equator> 
! 
! EOP
!
! Local:
      integer :: nint,j0,n,j,i
!
      nint = (nlon0-1)/(nlon-1)
      j0 = 1-nint
      do n = 1,9
        do j = 1,nlat-1
          do i = 1,nlon
            c(i,j,n) = c(i,j,n)/(cs(j0+j*nint)*nint**2)
          enddo ! i = 1,nlon
        enddo ! j = 1,nlat-1
      enddo ! n = 1,9
!
      end subroutine divide
!-----------------------------------------------------------------------
! BOP
! !IROUTINE: edges
! !INTERFACE:
      subroutine edges(c,nlon,nlat)
!
! !USES:
! !DESCRIPTION: 
! Insert polar boundary conditions in stencil c(nlon,nlat,9)
!    c(:,nlat,1:8) = 0 & diagonal term c(:,nlat,9) = 1.
!
! !ARGUMENTS:
      integer,intent(in) :: nlon,nlat    ! dimension of coeff. array
! !RETURN VALUE:
      real,intent(out) :: c(nlon,nlat,*) ! coefficient array
!        
! !REVISION HISTORY:
! 21.02.05  <Astrid Maute> <include header> 
! 
! EOP
!
! Local:
      integer :: n,i
!
      do n=1,8
        do i=1,nlon
          c(i,nlat,n) = 0.
        enddo
      enddo
      do i=1,nlon
        c(i,nlat,9) = 1.
      enddo
      end subroutine edges
!-----------------------------------------------------------------------
! BOP
! !IROUTINE: stenmod
! !INTERFACE:
      subroutine stenmod(inlon,inlat,c,phihm,pfrac)
!
! !USES:
      use cons_module,only: dlatm,dtr
!
! !DESCRIPTION:
! Modify stencil c and cofum to set potential to heelis value within auroral circle.
!    c     is the stencil w   upwinding
! pfrac set up in heelis module: fractional presence of dynamo equation
!        pfrac = 1  for |colam_m| < crit(2)
!        pfrac = 0  for |colam_m| > crit(1)
!        pfrac = (colat_m-crit(1))/(crit(2)-crit(1))  
!                                for crit(2) < |lam_m| < crit(1)
! to ensure diagonal dominance only the diogonal element c(9) is used
! factor of (dlatm/(10.*dtr))**2 to eliminate dependence on grid spacing
! for c(9)
!
! !PARAMETERS: 
      integer,intent(in) :: inlon,inlat
      real,dimension(nmlon0,nmlat0),intent(in) :: 
     |  phihm,  ! heelis potential (from subs potm, flwv32)
     |  pfrac   ! fractional presence of dynamo (from sub colath)
! !RETURN VALUE:
      real,intent(inout) :: c(inlon,inlat,*)
! !REVISION HISTORY:
! 05.03.10  <Astrid Maute> <include header> 
! 
! EOP
!
! Local:
      integer :: nint,i0,j0,i,j,n,jj
!
! Compute separation of grid points for this resolution:
      nint = (nmlon0-1)/(inlon-1)
      i0 = 1-nint
      j0 = 1-nint
!
! If nint==1, then we are at the highest resolution.
! Correct RHS, which is in c(10)
!
      if (nint==1) then
        do j=1,inlat
          do i=1,inlon
            c(i,j,10) = pfrac(i,j)*c(i,j,10)+(1.-pfrac(i,j))*c(i,j,9)*
     |        (dlatm/(10.*dtr))**2*phihm(i,j)
          enddo ! i=1,inlon
        enddo ! j=1,inlat
      endif
!
! Modify stencil, c(i,j,n),n=1,9:
!
      do j=1,inlat
        jj = j0+j*nint
        do n = 1,8
          do i = 1,inlon
            c(i,j,n) = c(i,j,n)*pfrac(i0+i*nint,jj)
          enddo ! i = 1,inlon
        enddo ! n = 1,8
        do i = 1,inlon
          c(i,j,9) = c(i,j,9)*pfrac(i0+i*nint,jj)+
     |      (1.-pfrac(i0+i*nint,jj))*c(i,j,9)*
     |      (dlatm*float(nint)/(10.*dtr))**2
        enddo ! i = 1,inlon
      enddo ! j=1,inlat
!
      end subroutine stenmod
!-----------------------------------------------------------------------
! BOP
! !IROUTINE: stenmd
! !INTERFACE:
      subroutine stenmd(inlon,inlat,c,phihm,pfrac)
! !USES:
      use cons_module,only: dlatm,dtr
!
! !DESCRIPTION:
! Modify stencil c and cofum to set potential to heelis value within auroral circle.
!    c     is the stencil w   upwinding
!    cofum is the stencil w/o upwinding
! pfrac set up in heelis module: fractional presence of dynamo equation
!        pfrac = 1  for |colam_m| < crit(2)
!        pfrac = 0  for |colam_m| > crit(1)
!        pfrac = (colat_m-crit(1))/(crit(2)-crit(1))  
!                                for crit(2) < |lam_m| < crit(1)
! to ensure diagonal dominance only the diogonal element c(9) is used
! factor of (dlatm/(10.*dtr))**2 to eliminate dependence on grid spacing
! for c(9)
!
! !PARAMETERS: 
      integer,intent(in) :: inlon,inlat
      real,dimension(nmlon0,nmlat0),intent(in) :: 
     |  phihm,  ! heelis potential (from subs potm, flwv32)
     |  pfrac   ! fractional presence of dynamo (from sub colath)
! !RETURN VALUE:
      real,intent(inout) :: c(inlon,inlat,*)
! !REVISION HISTORY:
! 05.03.8  <Astrid Maute> <include header> 
! 
! EOP
!
! Local:
      integer :: nint,i0,j0,i,j,n,jj
!
! Compute separation of grid points for this resolution:
      nint = (nmlon0-1)/(inlon-1)
      i0 = 1-nint
      j0 = 1-nint
!
! If nint==1, then we are at the highest resolution.
! Correct RHS, which is in c(10)
!
      if (nint==1) then
        do j=1,inlat
          do i=1,inlon
            c(i,j,10) = pfrac(i,j)*c(i,j,10)+(1.-pfrac(i,j))*c(i,j,9)*
     |        (dlatm/(10.*dtr))**2*phihm(i,j)
          enddo ! i=1,inlon
        enddo ! j=1,inlat
      endif
!
! Modify stencil, c(i,j,n),n=1,9:
!
      if (nint==1) then
        do j=1,inlat
          jj = j0+j*nint
          do n = 1,8
            do i = 1,inlon
              c(i,j,n) = c(i,j,n)*pfrac(i0+i*nint,jj)
              cofum_gip(i,j,n) = cofum_gip(i,j,n)*pfrac(i0+i*nint,jj)
            enddo ! i = 1,inlon
          enddo ! n = 1,8
          do i = 1,inlon
            c(i,j,9) = c(i,j,9)*pfrac(i0+i*nint,jj)+
     |        (1.-pfrac(i0+i*nint,jj))*c(i,j,9)*
     |        (dlatm*float(nint)/(10.*dtr))**2
            cofum_gip(i,j,9) =cofum_gip(i,j,9)*pfrac(i0+i*nint,jj)+
     |        (1.-pfrac(i0+i*nint,jj))*cofum_gip(i,j,9)*
     |        (dlatm*float(nint)/(10.*dtr))**2
          enddo ! i = 1,inlon
        enddo ! j=1,inlat
      else ! nint /= 1
        do j=1,inlat
          jj = j0+j*nint
          do n = 1,8
            do i = 1,inlon
              c(i,j,n) = c(i,j,n)*pfrac(i0+i*nint,jj)
            enddo ! i = 1,inlon
          enddo ! n = 1,8
          do i = 1,inlon
            c(i,j,9) = c(i,j,9)*pfrac(i0+i*nint,jj)+
     |        (1.-pfrac(i0+i*nint,jj))*c(i,j,9)*
     |        (dlatm*float(nint)/(10.*dtr))**2
          enddo ! i = 1,inlon
        enddo ! j=1,inlat
      endif ! nint
!      
      end subroutine stenmd
!-----------------------------------------------------------------------
! BOP
! !IROUTINE: threed
! !INTERFACE:
      subroutine threed
!
! !USES:
      use cons_module,only: ylatm,r0,r00,pi_dyn,table,dlatm,dlonm,
     |  rcos0s,re,dt1dts,dt0dts
      use fields_module,only: phim3d, ! phim3d(nmlonp1,nmlat,-2:nlevp1)
     |  emphi3d,   ! 3d eastward electric field magnetic
     |  emlam3d,   ! 3d equatorw. electric field magnetic
     |  emz3d      ! 3d upward (?) electric field magnetic
      use tgcm2gip_module,only: tgcm_ed1,tgcm_ed2
!
! !DESCRIPTION: 
! calculates electric field 
!      E_d1 = -1/(R cos lam_m) d Phi/ d phi_m
!      E_d2 = 1/ (R |sinI_m|)  d Phi/ d lam_m
! calculate 3D electric potential and electric field assuming
! that the fieldlines are dipolar and with constant electric potential
! along the field line
!
! !REVISION HISTORY:
! 07.03.05  <Astrid Maute> <include header> 
! 
! EOP
!
! Local:
      integer :: k,i,j,ip1f,ip2f,ip3f
      real :: cosltm,pi,phimn,phims,csth0,sym,ctm,sinltm,siniq
      real,dimension(nmlonp1) :: thetam,pslot,qslot
      integer,dimension(nmlonp1) :: islot,jslot
      real,dimension(nmlonp1,nmlat) :: ephi,elam
!
! Externals:
      real,external :: sddot ! in util.F
!
      pi = pi_dyn
!
! Calculate E_d1, E_d2 components of electric field:
! E_d1 = -1/(R cos lam_0)* (cos lam_0/cos lam_m) d Phi/ d phi_m
      do j=2,nmlat-1
        csth0 = cos(-pi/2.+(j-1)*dlatm)
        do i=2,nmlon
          ed1(i,j) = -(phim(i+1,j)-phim(i-1,j))/(2.*dlonm*csth0)*
     |      rcos0s(j)/(r0*1.e-2)
          ephi(i,j) = ed1(i,j)*(r0*1.e-2)
        enddo ! i=2,nmlon
        ed1(1,j) = -(phim(2,j)-phim(nmlon,j))/(2.*dlonm*csth0)*
     |      rcos0s(j)/(r0*1.e-2)
        ed1(nmlonp1,j) = ed1(1,j)
        ephi(1,j)       = ed1(1,j)*(r0*1.e-2)
        ephi(nmlonp1,j) = ed1(nmlonp1,j)*(r0*1.e-2)
      enddo ! j=2,nmlat-1
!
! E_d2 = +/- 1/(R |sinI_m|) d Phi/d lam_0* (d lam_0/ (d lam_m*|sin I_m|))
! Southern hemisphere:
      do j=2,nmlath-1
        do i=1,nmlonp1
          ed2(i,j) = -(phim(i,j+1)-phim(i,j-1))/(2.*dlatm)*dt1dts(j)/
     |      (r0*1.e-2)
          elam(i,j) = -(phim(i,j+1)-phim(i,j-1))/(2.*dlatm)*dt0dts(j)
        enddo ! i=1,nmlonp1
      enddo ! j=2,nmlath-1
!
! Northern hemisphere:
      do j=nmlath+1,nmlat-1
        do i=1,nmlonp1
          ed2(i,j) = (phim(i,j+1)-phim(i,j-1))/(2.*dlatm)*dt1dts(j)/
     |      (r0*1.e-2)
          elam(i,j) = -(phim(i,j+1)-phim(i,j-1))/(2.*dlatm)*dt0dts(j)
        enddo ! i=1,nmlonp1
      enddo ! j=nmlath+1,nmlat-1
!
! Poles:
      do i = 1,nmlonp1
        ip1f = i + nmlon/4
        if (ip1f > nmlonp1) ip1f = ip1f - nmlon
        ip2f = i + nmlon/2
        if (ip2f > nmlonp1) ip2f = ip2f - nmlon
        ip3f = i + 3*nmlon/4
        if (ip3f > nmlonp1) ip3f = ip3f - nmlon
        ed1(i,1) = .25*(ed1(i,2) - ed1(ip2f,2) +
     |                  ed2(ip1f,2) - ed2(ip3f,2))
        ed1(i,nmlat) = .25*(ed1(i,nmlat-1) - ed1(ip2f,nmlat-1) +
     |                      ed2(ip1f,nmlat-1) - ed2(ip3f,nmlat-1))
        ed2(i,1) = .25*(ed2(i,2) - ed2(ip2f,2) -
     |                  ed1(ip1f,2) + ed1(ip3f,2))
        ed2(i,nmlat) = .25*(ed2(i,nmlat-1) - ed2(ip2f,nmlat-1) -
     |                      ed1(ip1f,nmlat-1) + ed1(ip3f,nmlat-1))
        ephi(i,1) = .25*(ephi(i,2) - ephi(ip2f,2) +
     |                  elam(ip1f,2) - elam(ip3f,2))
        ephi(i,nmlat) = .25*(ephi(i,nmlat-1) - ephi(ip2f,nmlat-1) +
     |                      elam(ip1f,nmlat-1) - elam(ip3f,nmlat-1))
        elam(i,1) = .25*(elam(i,2) - elam(ip2f,2) -
     |                  ephi(ip1f,2) + ephi(ip3f,2))
        elam(i,nmlat) = .25*(elam(i,nmlat-1) - elam(ip2f,nmlat-1) -
     |                      ephi(ip1f,nmlat-1) + ephi(ip3f,nmlat-1))
!
! Equator
       ed2(i,nmlath) = (4.*phim(i,nmlath+1)-phim(i,nmlath+2)
     |    -3.*phim(i,nmlath))/(2.*dlatm)*dt1dts(nmlath)/(R0*1.e-2)
       ed2(i,nmlath+1) = (4.*phim(i,nmlath+2)-phim(i,nmlath+3)
     |    -3.*phim(i,nmlath+1))/(2.*dlatm)*dt1dts(nmlath+1)/(R0*1.e-2)
       ed2(i,nmlath-1) = (4.*phim(i,nmlath-2)-phim(i,nmlath-3)
     |    -3.*phim(i,nmlath-1))/(2.*dlatm)*dt1dts(nmlath-1)/(R0*1.e-2)
        elam(i,nmlath) = (4.*phim(i,nmlath+1)-phim(i,nmlath+2)
     |    -3.*phim(i,nmlath))/(2.*dlatm)
      enddo ! i = 1,nmlonp1
      
      call addfld('ED1','ED1',' ',ed1,'mlon',1,nmlonp1,
     |    'mlat',1,nmlat,0)
      call addfld('ED2','ED2',' ',ed2,'mlon',1,nmlonp1,
     |    'mlat',1,nmlat,0)
      
       tgcm_ed1 = ed1
       tgcm_ed2 = ed2
!
! Comment from tgcm15 (threed.F):
! For each geomagnetic grid point over the 3d grid, determine the
!   location of the foot of the magnetic field passing through the
!   grid point. Assume that the field lines are dipole in shape. 
!   Recalling that the model assumes constant electric potential
!   along each field line, interpolation in the 2d array phim gives
!   the potential at the original grid point.
! ar 3/2010
! For calculating electric fields and ExB drifts as a function of
!  height, neglect height variations of R_E+h, except when those height
!  variations cause large height variations in functions like
!  arccos{sqrt[(R_E+h_0)/(R_E+h)]}.  Consistent with this assumption,
!  neglect height variations of the geomagnetic-field strength.
!
      do k=-2,nlevp1
        do j=2,nmlat-1
          cosltm = cos(ylatm(j))     ! dipole latitude        
	  sinltm = sin(ylatm(j))
          sym = 1.                   ! northern hemisphere
          if (j < nmlath) sym = -1.  ! southern hemisphere
          do i=1,nmlon
!
! latitude lam_m of footpoint
! cos lam_m = cos lam_q [(R_E+h_0)/R_E+h]^0.5 
!
            ctm = sqrt(r0/(re+zpotenm3d(i,j,k)))*cosltm 
            thetam(i) = acos(ctm)
	    siniq = 2.*sinltm/sqrt(4.-3.*cosltm**2)
!
! Find position of lam_m in table of lam_m vs lam_0
! table is spaced 1 deg in lam_m
! Determine lam_0 (theta0) by interpolation in table:
            pslot(i) = thetam(i)*180./pi+1.
            islot(i) = int(pslot(i))
            pslot(i) = pslot(i)-float(islot(i))
            thetam(i) = ((1.-pslot(i))*table(islot(i),2)+pslot(i)*
     |        table(islot(i)+1,2))*sym ! thetam negative for south hem
!
! Now identify geomagnetic grid element in which the point lies:
! longitudinal grid point are the same
            islot(i) = i
            qslot(i) = (thetam(i)+pi/2.)/dlatm+1.
            jslot(i) = qslot(i)
            qslot(i) = qslot(i)-float(jslot(i))
!
! Interpolate in phim using linear interpolation in latitude:
! no interpolation necessary in longitude
            phim3d(i,j,k) = (1.-qslot(i))*
     |        phim(islot(i)  ,jslot(i)) +qslot(i)*
     |        phim(islot(i)  ,jslot(i)+1)
!
            ed13d(i,j,k) = (1.-qslot(i))*
     |        ed1(islot(i),jslot(i))+qslot(i)*
     |        ed1(islot(i),jslot(i)+1)
!
            ed23d(i,j,k) = (1.-qslot(i))*
     |        ed2(islot(i),jslot(i))+qslot(i)*
     |        ed2(islot(i),jslot(i)+1)
!
            emphi3d(i,j,k) = (1.-qslot(i))*
     |        ephi(islot(i),jslot(i))+qslot(i)*
     |        ephi(islot(i),jslot(i)+1)
!	    
            emlam3d(i,j,k) = -siniq*(r0*1.e-2)*( (1.-qslot(i))*
     |        ed2(islot(i),jslot(i))+qslot(i)*
     |        ed2(islot(i),jslot(i)+1) )
          enddo ! i=1,nmlon
        enddo ! j=2,nmlat-1
      enddo ! k=-2,nlevp1
!
! Calculate value of phim3d at geomagnetic poles:
      phims = sddot(nmlon,unitvm,phim(1,1))/float(nmlon)
      phimn = sddot(nmlon,unitvm,phim(1,nmlat))/float(nmlon)
      do k = -2,nlevp1
        do i = 1,nmlon
          phim3d(i,1,k) = phims
          phim3d(i,nmlat,k) = phimn
!
          ed13d(i,1,k)     = ed1(i,1)
          ed13d(i,nmlat,k) = ed1(i,nmlat)
          ed23d(i,1,k)     = ed2(i,1)
          ed23d(i,nmlat,k) = ed2(i,nmlat)
!
          emphi3d(i,1,k)     = ephi(i,1)
          emphi3d(i,nmlat,k) = ephi(i,nmlat)    
	  emlam3d(i,1,k)     =  ed2(i,1)*(r0*1.e-2)
	  emlam3d(i,nmlat,k) = -ed2(i,nmlat)*(r0*1.e-2)
        enddo ! i = 1,nmlon
      enddo ! k = -2,nlevp1
!          
      do k=2,nlevp1-1
        do j=1,nmlat
          do i=1,nmlon
             emz3d(i,j,k) = -(phim3d(i,j,k+1)-phim3d(i,j,k-1))
          enddo ! i=1,nmlon
        enddo ! j=2,nmlat-1
      enddo ! k=1,nlevp1  
!
      do k=-1,nlev
        do j=1,nmlat
          do i=1,nmlon
             emz3d(i,j,k) = -(phim3d(i,j,k+1)-phim3d(i,j,k-1))
          enddo ! i=1,nmlon
        enddo ! j=2,nmlat-1
      enddo ! k=1,nlevp1
!
! Periodic point:
      do k=1,nlevp1
        do j=1,nmlat
          phim3d(nmlonp1,j,k) = phim3d(1,j,k)
          ed13d(nmlonp1,j,k)  = ed13d(1,j,k)
          ed23d(nmlonp1,j,k)  = ed23d(1,j,k)
          emphi3d(nmlonp1,j,k)= emphi3d(1,j,k)
          emlam3d(nmlonp1,j,k)= emlam3d(1,j,k)
          emz3d(nmlonp1,j,k)  = emz3d(1,j,k)
        enddo ! j=1,nmlat
      enddo ! k=1,nlevp1
!
! Save 3d electric potential on magnetic grid to secondary history:
      do j=1,nmlat
  	call addfld('EMPHI3D','R E_mphi','[V/m*m]',
     |   emphi3d(:,j,:),'mlon',1,nmlonp1,'imlev',1,nmlev+1,j)
  	call addfld('EMLAM3D','R E_mlam','[V/m*m]',
     |   emlam3d(:,j,:),'mlon',1,nmlonp1,'imlev',1,nmlev+1,j)
  	call addfld('EMZ3D','E_mZ','[V/m*m]',
     |   emz3d(:,j,:),'mlon',1,nmlonp1,'imlev',1,nmlev+1,j)
       call addfld('ED1M3D','ed_1 (magn.)','[V/m]',
     |   ed13d(:,j,:),'mlon',1,nmlonp1,'imlev',1,nmlev+1,j)
       call addfld('ED2M3D'  ,'ed_2 (magn.)','[V/m]',
     |   ed23d(:,j,:),'mlon',1,nmlonp1,'imlev',1,nmlev+1,j)
       call addfld('PHIM3D','El.Poten','V',
     |    phim3d(:,j,:),'mlon',1,nmlonp1,'imlev',1,nmlev+1,j)
       call addfld('ZPOTEN3D','ZPOTEN3D','[cm]',
     |    zpotenm3d(:,j,:),'mlon',1,nmlonp1,'imlev',1,nmlev+1,j)
      enddo ! j=1,nmlat
!      
!
      end subroutine threed
!-----------------------------------------------------------------------
! BOP
! !IROUTINE: mag2geo
! !INTERFACE:
      subroutine mag2geo(am,ag,im,jm,dim,djm,lg,lm,nlong,nlatg,nlonm,
     |  nlatm)
! !USES:
! !DESCRIPTION:
! Transform field am on geomagnetic grid to ag on the geographic grid using
!   indices im,jm and weights am. Return field ag on geographic grid.
!
! !PARAMETERS: 
      integer,intent(in) :: lg,lm,nlong,nlatg,nlonm,nlatm
      integer,intent(in) :: im(lg,*),jm(lg,*)
      real,intent(in) :: am(lm,*),dim(lg,*),djm(lg,*)
! !RETURN VALUE:
      real,intent(out) :: ag(lg,*)
!
! !REVISION HISTORY:
! 05.03.11  <Astrid Maute> <include header> 
! 
! EOP
!
! Local:
      integer :: ig,jg
!
      do jg=1,nlatg 
        do ig=1,nlong
          ag(ig,jg) =
     |    am(im(ig,jg)  ,jm(ig,jg))  *(1.-dim(ig,jg))*(1.-djm(ig,jg))+
     |    am(im(ig,jg)+1,jm(ig,jg))  *    dim(ig,jg) *(1.-djm(ig,jg))+
     |    am(im(ig,jg)  ,jm(ig,jg)+1)*(1.-dim(ig,jg))*djm(ig,jg)+
     |    am(im(ig,jg)+1,jm(ig,jg)+1)*    dim(ig,jg) *djm(ig,jg)
        enddo ! ig=1,nlong
      enddo ! jg=1,nlatg 
      end subroutine mag2geo
!-----------------------------------------------------------------------
! BOP
! !IROUTINE: geo2mag
! !INTERFACE:
      subroutine geo2mag(fmag,fgeo,long,latg,wght,nlonp1_geo,nlonp1_mag,
     |  nlon_mag,nlat_mag,lat)
! !USES:
!
! !DESCRIPTION:
! Transform field fgeo on geographic grid to geomagnetic grid using
!   indices long,latg and weights wght. Return field fmag on magnetic grid.
!
! !PARAMETERS: 
      integer,intent(in) :: nlonp1_geo,nlonp1_mag,nlon_mag,nlat_mag,lat
      integer,dimension(nlonp1_mag,nlat_mag),intent(in) :: long,latg
      real,intent(in) :: fgeo(nlonp1_geo,*),wght(4,nlonp1_mag,nlat_mag)
! !RETURN VALUE:
      real,intent(out) :: fmag(nlonp1_mag,*)
!
! !REVISION HISTORY:
! 05.03.8  <Astrid Maute> <include header> 
! 
! EOP
!
! Local:
      integer :: i
!
      do i=1,nlon_mag
        fmag(i,1) =
     |    fgeo(long(i,lat)  ,latg(i,lat)  )*wght(1,i,lat)+
     |    fgeo(long(i,lat)+1,latg(i,lat)  )*wght(2,i,lat)+
     |    fgeo(long(i,lat)+1,latg(i,lat)+1)*wght(3,i,lat)+
     |    fgeo(long(i,lat)  ,latg(i,lat)+1)*wght(4,i,lat)
!       if (iprint > 0) write(6,"('geo2mag: i=',i3,' lat=',i3,' long=',
!    |    i3,' latg=',i3,' wght=',4e12.4,' fgeo=',e12.4,' fmag=',
!    |    e12.4)") i,lat,long(i,lat),latg(i,lat),wght(:,i,lat),
!    |    fgeo(long(i,lat),latg(i,lat)),fmag(i,1)
      enddo
      end subroutine geo2mag
!-----------------------------------------------------------------------
! BOP
! !IROUTINE: prep_dynamo_gip
! !INTERFACE:
      subroutine prep_dynamo_gip(z,lev0,lev1,lon0,lon1,lat0,lat1)
!      
! !USES:
      use cons_module,only: gask,grav
!      
! !DESCRIPTION:
! am_11/03: if dynamo = 0 need zpoten3d for subroutine threed
! Prepare geographic-grid fields for input to the dynamo, and gather them
! to the root task. This is executed by all tasks, and is called from 
! advance before the dynamo itself (which is executed by master task only).
!
! !PARAMETERS: 
      integer :: lev0,lev1,lon0,lon1,lat0,lat1
!
! Input fields at task subdomains. These are from current (not updated)
! fields (itp):
      real,dimension(lev0:lev1,lon0-2:lon1+2,lat0-2:lat1+2),intent(in)::
     |  z     ! geopotential height
!     
! !REVISION HISTORY:
! 05.03.8  <Astrid Maute> <include header> 
! 
! EOP
!
! Local:
      integer :: k,i,lat,nk,lonbeg,lonend
      real :: fmin,fmax
!
      nk = lev1-lev0+1
      lonbeg = lon0
      if (lon0==1) lonbeg = 3
      lonend = lon1
      if (lon1==nlonp4) lonend = nlonp4-1
!
! Define subdomain part of dynamo input fields (geographic):
! Also transform from (k,i,lat) to (i,lat,k). The latter is used
!   in the dynamo code, whereas the former is used in the rest
!   of the model.
! (In earlier versions, this was in lamdas, where 1-73 <= 3-75.
!  Here, 3-75 <= 3-75)
!
      do lat=lat0,lat1
        do i=lonbeg,lonend
          do k=lev0,lev1-1
            zpoten    (i,lat,k) = z   (k,i,lat)
          enddo ! k=lev0,lev1
          zpoten(i,lat,lev1) = z(lev1,i,lat)
        enddo ! i=lon0,lon1
!
      enddo ! lat=lat0,lat1

#ifdef MPI
!
! Gather dynamo input fields to the root task, defining module data 
! above at the global domain on the root task:
!
!      call mp_dynamo_gather
#endif
      end subroutine prep_dynamo_gip
!-----------------------------------------------------------------------
! BOP
! !IROUTINE: set_zpoten
! !INTERFACE:
      subroutine set_zpoten
!
! !USES:
      use cons_module,only: h0,r0
      use magfield_module,only: 
     |  zb,       ! (nlonp1,0:nlatp1)
     |  ig,       ! (nmlonp1,nmlat) geog lon at each geomag grid point
     |  jg,       ! (nmlonp1,nmlat) geog lat at each geomag grid point
     |  wt        ! (4,nmlonp1,nmlat) interpolation weights
!     
! !RETURN VALUE:
! !PARAMETERS: 
! !DESCRIPTION:
! am_11/03  used only if dynamo = 0
! need zpoten3d in magnetic coordinates for subroutine threed
! !REVISION HISTORY:
! 05.03.8  <Astrid Maute> <include header> 
! 06.03.31 copied from transf_dyn0
! 
! EOP
!
! Local:
      integer :: i,ii,k,kk,j,jj,lat,n
      real :: z0
!
! Fields to be transformed to geomagnetic space (formerly in transmag.h):
! (these will be input to the geographic to magnetic transformation). 
      real,dimension(nlonp1,0:nlatp1,-2:nlevp1) :: zz
!
      real,dimension(nlonp4,nlevp1) :: zz_plt
!
! External:
      real,external :: sddot ! in util.F
!
! Set constants:
!   z0 is lowest level for start of field line integration set in h0 
!     (h0 in cons module)
!
!
! Pack inputs from 3->nlon+3 to 1->nlon+1, as in tgcm15 (end of sub lamdas):
      do lat=1,nlat
        do i=1,nlon+1
          ii = i+2
          zpoten(i,lat,:) = zpoten(ii,lat,:)
        enddo ! i=1,nlonp4-3
      enddo ! lat=1,nlat
!
      z0 = h0
!
! Calculate quantities to be transformed to geomagnetic space:
      do k=1,nlev        
        do j=1,nlat
          do i=1,nlonp1
            zz(i,j,k) = zpoten(i,j,k)
          enddo ! i=1,nlonp1
        enddo ! j=1,nlat
      enddo ! k=1,nlev        
      do j=1,nlat
        do i=1,nlonp1
          zz(i,j,nlevp1) = zpoten(i,j,nlevp1)
        enddo ! i=1,nlonp1
      enddo ! j=1,nlat
!
! Extend fields down to 90 km inserting 3 extra levels. 
! Set three equally spaced levels for Z, take U, V, and W
!   to be constant, and extrapolate sigmas exponentially.
!
      do k=0,-2,-1
        do j=1,nlat
          do i=1,nlonp1
            zz(i,j,k) = z0+float(k+2)*(zz(i,j,1)-z0)/3.
          enddo ! i=1,nlonp1
        enddo ! j=1,nlat
      enddo ! k=0,-2,-1
!
! Values at poles:
      do k=-2,nlev
        zz(1,0,k) = 
     |    (9.*sddot(nlon,unitv,zz(1,1,k))-
     |        sddot(nlon,unitv,zz(1,2,k)))/(8.*float(nlon))
        zz(1,nlatp1,k) = 
     |    (9.*sddot(nlon,unitv,zz(1,nlat  ,k))-
     |        sddot(nlon,unitv,zz(1,nlat-1,k)))/(8.*float(nlon))
!
! Extend in longitude:
        do i = 2,nlon
          zz(i,0,k)             = zz(1,0,k)
          zz(i,nlatp1,k)        = zz(1,nlatp1,k)
        enddo ! i = 2,nlon
      enddo ! k=-2,nlev
!
! Values at the poles:
      zz(1,0,nlevp1)= (9.*sddot(nlon,unitv,zz(1,1,nlevp1))-
     |                  sddot(nlon,unitv,zz(1,2,nlevp1)))/
     |                  (8.*float(nlon))
      zz(1,nlatp1,nlevp1) =  (9.*
     |                  sddot(nlon,unitv,zz(1,nlat,nlevp1))-
     |                  sddot(nlon,unitv,zz(1,nlat-1,nlevp1)))/
     |                  (8.*float(nlon))
!
! Extend in longitude:
      do i = 2,nlon
        zz(i,0,nlevp1)     = zz(1,0,nlevp1)
        zz(i,nlatp1,nlevp1)= zz(1,nlatp1,nlevp1)
      enddo ! i = 2,nlon
!
! Periodic points:
      do j = 0,nlatp1
        do k = -2,nlev
           zz(nlonp1,j,k)      = zz(1,j,k)
        enddo ! k = -2,nlev
        zz(nlonp1,j,nlevp1)= zz(1,j,nlevp1)
      enddo ! j = 0,nlatp1
!
! Transform needed fields to geomagnetic coordinate system, one latitude
! at a time. 
!
! subroutine geo2mag(fmag,fgeo,long,latg,wght,nlonp1_geo,nlonp1_mag,
!|  nlon_mag,nlat_mag,lat)
!
! Magnetic latitude loop:
      maglat_loop: do j=2,nmlat-1
        do k=-2,nlevp1
          call geo2mag(zpotenm(1,k),zz(1,0,k),ig,jg,wt,nlonp1,
     |      nmlonp1,nmlon,nmlat,j)
        enddo ! k=-2,nlevp1
! Without calculation of k_(q,lam)
        do i=1,nmlon  
          do k=-2,nlev
            if (zpotenm(i,k) < z0) zpotenm(i,k) = z0
            zpotenm3d(i,j,k) = zpotenm(i,k)
          enddo ! k=-2,nlev
          if (zpotenm(i,nlevp1) < z0) zpotenm(i,nlevp1) = z0
          zpotenm3d(i,j,nlevp1) = zpotenm(i,nlevp1)
        enddo ! i=1,nmlon 
      enddo maglat_loop ! j=2,nmlat-1 Main magnetic latitude loop
!
! Polar values for Z:
      do k=-2,nlevp1
        zpotenm3d(1,1,k) = 
     |    (4.*sddot(nmlon,unitvm,zpotenm3d(1,2,k))-
     |        sddot(nmlon,unitvm,zpotenm3d(1,3,k)))/(3.*float(nmlon))
        zpotenm3d(1,nmlat,k) = 
     |    (4.*sddot(nmlon,unitvm,zpotenm3d(1,nmlat-1,k))-
     |        sddot(nmlon,unitvm,zpotenm3d(1,nmlat-2,k)))/
     |    (3.*float(nmlon))
!
! Extend Z over longitude:
        do i=1,nmlon
          zpotenm3d(i,1,k) = zpotenm3d(1,1,k)
          zpotenm3d(i,nmlat,k) = zpotenm3d(1,nmlat,k)
        enddo ! i=1,nmlon
!
! Periodic points:
        do j=1,nmlat
          zpotenm3d(nmlonp1,j,k) = zpotenm3d(1,j,k)
        enddo ! j=1,nmlat
      enddo ! k=-2,nlevp1
!
! Save 3d potential on magnetic grid to secondary history:
! PLEASE DO NOT COMMENT THIS OUT -- ZMAG is a mandatory mag field for sech
       do j=1,nmlat
!         call addfsech_ik('ZMAG','ZMAG','CM',zpotenm3d(:,j,:),
!     | 	   1,nmlonp1,nmlev,nmlev,j)
       enddo ! j=1,nmlat
!
      end subroutine set_zpoten
!-----------------------------------------------------------------------
      end module dynamo_gip_module
!-----------------------------------------------------------------------
